from abc import ABC, abstractmethod

from flloat.semantics.pl import PLInterpretation


class Delta(ABC):

    @abstractmethod
    def delta(self, i:PLInterpretation, epsilon=False):
        raise NotImplementedError



from abc import abstractmethod
from typing import List, Set

from flloat.base.Interpretation import Interpretation
from flloat.base.Symbol import Symbol
from flloat.base.Symbols import Symbols
from flloat.base.truths import Truth
from flloat.semantics.pl import PLInterpretation



class FiniteTrace(Interpretation):
    def __init__(self, trace: List[PLInterpretation]):
        super().__init__()
        self.trace = trace

    def _members(self):
        return tuple(self.trace)

    @staticmethod
    def fromSymbolSets(l:List[Set[Symbol]]):
        return FiniteTrace([PLInterpretation(s) for s in l])

    @staticmethod
    def fromStringSets(l:List[Set[str]]):
        return FiniteTrace([PLInterpretation(frozenset({Symbol(string) for string in s})) for s in l])

    def length(self):
        return len(self.trace)

    def last(self):
        return len(self.trace)-1

    def _position_is_legal(self, position:int):
        return position>=0 and position <= self.last()

    def get(self, position:int) -> PLInterpretation:
        assert self._position_is_legal(position)
        return self.trace[position]

    def segment(self, start:int, end:int) :
        if not self._position_is_legal(start) or not self._position_is_legal(end):
            raise ValueError("Start or end position are not valid")
        return FiniteTrace(self.trace[start: end])

    def __str__(self):
        return "Trace (length=%s)" %self.length() + "\n\t" + \
               "\n\t".join("%d: {"%i + ", ".join(map(str,sorted(e))) + "}" for i, e in enumerate(self.trace))


class FiniteTraceTruth(Truth):
    @abstractmethod
    def truth(self, i: FiniteTrace, pos: int):
        raise NotImplementedError



from abc import abstractmethod
from functools import lru_cache
from typing import Set

from flloat.base.Formula import Formula, CommutativeBinaryOperator, AtomicFormula, BinaryOperator, UnaryOperator
from flloat.base.Symbol import Symbol
from flloat.base.Symbols import Symbols
from flloat.base.convertible import ImpliesDeltaConvertible, EquivalenceDeltaConvertible, ConvertibleFormula, \
    BaseConvertibleFormula, DeltaConvertibleFormula
from flloat.base.nnf import NNF, NotNNF, DualBinaryOperatorNNF, DualUnaryOperatorNNF
from flloat.base.truths import Truth, NotTruth, OrTruth, AndTruth
from flloat.utils import MAX_CACHE_SIZE
from flloat.flloat import DFAOTF, to_automaton, to_automaton_
from flloat.semantics.ldlf import FiniteTrace, FiniteTraceTruth
from flloat.semantics.pl import PLInterpretation, PLFalseInterpretation
from flloat.syntax.ldlf import Delta, LDLfAtomic, LDLfNot, LDLfAnd, LDLfOr, LDLfEquivalence, LDLfDiamond, \
    RegExpPropositional, RegExpStar, RegExpSequence, RegExpTest, LDLfPropositional, LDLfBox, LDLfEnd
from flloat.syntax.pl import PLTrue, PLFalse, PLAnd, PLOr, PLAtomic


class LTLfTruth(Truth):
    @abstractmethod
    def truth(self, i: FiniteTrace, pos: int = 0):
        raise NotImplementedError


class LTLfFormula(Formula, LTLfTruth, NNF, Delta):

    @lru_cache(maxsize=MAX_CACHE_SIZE)
    def delta(self, i: PLInterpretation, epsilon=False):
        f = self.to_nnf()
        # f = self.to_LDLf().to_nnf()
        d = f._delta(i, epsilon)
        if epsilon:
            # By definition, if epsilon=True, then the result must be either PLTrue or PLFalse
            # Now, the output is a Propositional Formula with only PLTrue or PLFalse as atomics
            # Hence, we just evaluate the formula with a dummy PLInterpretation
            d = PLTrue() if d.truth(None) else PLFalse()
        return d

    @abstractmethod
    def _delta(self, i: PLInterpretation, epsilon=False):
        """apply delta function, assuming that 'self' is a LTLf formula in Negative Normal Form"""
        raise NotImplementedError

    @abstractmethod
    def to_LDLf(self):
        raise NotImplementedError

    def __repr__(self):
        return self.__str__()

    def to_automaton(self, labels:Set[Symbol]=None, on_the_fly=False, determinize=False, minimize=True):
        if labels is None:
            labels = self.find_labels()
        if on_the_fly:
            return DFAOTF(self)
        elif determinize:
            return to_automaton(self, labels, minimize)
        else:
            return to_automaton_(self, labels)


class LTLfCommBinaryOperator(CommutativeBinaryOperator, LTLfFormula):
    pass

class LTLfTemporalFormula(LTLfFormula, FiniteTraceTruth):
    pass


class LTLfAtomic(AtomicFormula, LTLfFormula):

    def __init__(self, s:Symbol):
        AtomicFormula.__init__(self, s)
        self.a = PLAtomic(s)

    def __str__(self):
        return AtomicFormula.__str__(self)

    def _members(self):
        return (LTLfAtomic, self.s)

    def _to_nnf(self):
        return self

    def negate(self):
        return LTLfNot(LTLfAtomic(self.s))

    def _delta(self, i:PLInterpretation, epsilon=False):
        if epsilon:
            return PLFalse()
        return PLTrue() if self.a.truth(i) else PLFalse()

    def truth(self, i: FiniteTrace, pos: int=0):
        return self.a.truth(i.get(pos))

    def find_labels(self):
        return self.a.find_labels()

    def to_LDLf(self):
        return LDLfPropositional(self.a)._convert()


class LTLfTrue(LTLfAtomic):
    def __init__(self):
        super().__init__(Symbol(Symbols.TRUE.value))
        self.a = PLTrue()

    def negate(self):
        return LTLfFalse()


class LTLfFalse(LTLfAtomic):
    def __init__(self):
        super().__init__(Symbol(Symbols.FALSE.value))
        self.a = PLFalse()

    def negate(self):
        return LTLfTrue()



class LTLfNot(NotTruth, LTLfFormula, NotNNF):

    def _delta(self, i: PLInterpretation, epsilon=False):
        if isinstance(self.f, LTLfAtomic) or isinstance(self.f, LTLfEnd):
            return PLTrue() if self.f._delta(i, epsilon)==PLFalse() else PLFalse()
        else:
            # the formula must be in NNF form!!!
            raise Exception

    def to_LDLf(self):
        return LDLfNot(self.f.to_LDLf())

class LTLfAnd(LTLfCommBinaryOperator, AndTruth, DualBinaryOperatorNNF):

    def _delta(self, i:PLInterpretation, epsilon=False):
        return PLAnd([f._delta(i, epsilon) for f in self.formulas])

    def to_LDLf(self):
        return LDLfAnd([f.to_LDLf() for f in self.formulas])


class LTLfOr(LTLfCommBinaryOperator, OrTruth, DualBinaryOperatorNNF):

    def _delta(self, i:PLInterpretation, epsilon=False):
        return PLOr([f._delta(i, epsilon) for f in self.formulas])

    def to_LDLf(self):
        return LDLfOr([f.to_LDLf() for f in self.formulas])


class LTLfImplies(ImpliesDeltaConvertible, LTLfFormula):
    And = LTLfAnd
    Or = LTLfOr
    Not = LTLfNot

    def to_LDLf(self):
        return self._convert().to_LDLf()

class LTLfEquivalence(EquivalenceDeltaConvertible, LTLfCommBinaryOperator):
    And = LTLfAnd
    Or = LTLfOr
    Not = LTLfNot

    def to_LDLf(self):
        return self._convert().to_LDLf()


class LTLfNext(DualUnaryOperatorNNF, LTLfTemporalFormula):
    operator_symbol = "X"
    Not = LTLfNot

    def truth(self, i: FiniteTrace, pos: int=0):
        return pos < i.last() and self.f.truth(i, pos + 1)

    def _delta(self, i: PLInterpretation, epsilon=False):
        if epsilon:
            return PLFalse()
        else:
            return PLAnd([self.f, LTLfNot(LTLfEnd()).to_nnf()])

    def to_LDLf(self):
        return LDLfDiamond(RegExpPropositional(PLTrue()), LDLfAnd([self.f.to_LDLf(), LDLfNot(LDLfEnd())]))


class LTLfWeakNext(DualUnaryOperatorNNF, ConvertibleFormula, LTLfTemporalFormula):
    operator_symbol = Symbols.WEAK_NEXT.value
    Not = LTLfNot

    def _convert(self):
        return LTLfNot(LTLfNext(LTLfNot(self.f)))

    def truth(self, i: FiniteTrace, pos: int=0):
        return self._convert().truth(i, pos)

    def _delta(self, i: PLInterpretation, epsilon=False):
        if epsilon:
            return PLTrue()
        else:
            return PLOr([self.f, LTLfEnd().to_nnf()])
            # return LTLfOr([self.f, LTLfWeakNext(LTLfFalse())])

            # if self.f == LTLfFalse():
            #     return LTLfFalse()
            # return PLOr([self.f, LTLfWeakNext(LTLfFalse())])

    def to_LDLf(self):
        return self._convert().to_LDLf()


class LTLfUntil(DualBinaryOperatorNNF, LTLfTemporalFormula):
    operator_symbol = "U"

    def _to_nnf(self):
        return LTLfUntil([f.to_nnf() for f in self.formulas])

    def negate(self):
        return LTLfRelease([LTLfNot(f) for f in self.formulas])

    def truth(self, i: FiniteTrace, pos: int=0):
        f1 = self.formulas[0]
        f2 = LTLfUntil(self.formulas[1:]) if len(self.formulas)>2 else self.formulas[1]

        a = any(f2.truth(i, j) for j in range(pos, i.last()+1))
        b = all(f1.truth(i, k) for j in range(pos, i.last()+1) for k in range(pos, j))

        return any(f2.truth(i, j) and all(f1.truth(i, k) for k in range(pos, j)) for j in range(pos, i.last()+1))

    def _delta(self, i:PLInterpretation, epsilon=False):
        if epsilon:
            return PLFalse()
        f1 = self.formulas[0]
        f2 = LTLfUntil(self.formulas[1:]) if len(self.formulas) > 2 else self.formulas[1]
        return PLOr([
            f2._delta(i, epsilon),
            PLAnd([
                f1._delta(i, epsilon),
                LTLfNext(self)._delta(i, epsilon)
            ])
        ])

    def to_LDLf(self):
        f1 = self.formulas[0].to_LDLf()
        f2 = LTLfUntil(self.formulas[1:]).to_LDLf() if len(self.formulas) > 2 else self.formulas[1].to_LDLf()
        return LDLfDiamond(RegExpStar(RegExpSequence([RegExpTest(f1), RegExpPropositional(PLTrue())])), LDLfAnd([f2, LDLfNot(LDLfEnd())]))


class LTLfEventually(UnaryOperator, DeltaConvertibleFormula, LTLfTemporalFormula):
    operator_symbol = "F"
    Not = LTLfNot

    def _convert(self):
        return LTLfUntil([LTLfTrue(), self.f])

    # def _delta(self, i:PLInterpretation, epsilon=False):
        # if epsilon:
        #     return PLFalse()
        # else:
        #     return PLOr([self.f._delta(i, epsilon), LTLfNext(self)._delta(i, epsilon)])


    def to_LDLf(self):
        # return self._convert().to_LDLf()
        return LDLfDiamond(RegExpStar(RegExpPropositional(PLTrue())), LDLfAnd([self.f.to_LDLf(), LDLfNot(LDLfEnd())]))


class LTLfAlways(UnaryOperator, DeltaConvertibleFormula, LTLfTemporalFormula):
    operator_symbol = "G"
    Not = LTLfNot

    def _convert(self):
        return LTLfNot(LTLfEventually(LTLfNot(self.f))._convert())

    # def _delta(self, i:PLInterpretation, epsilon=False):
    #     if epsilon:
    #         return PLTrue()
    #     else:
    #         return PLAnd([self.f._delta(i, epsilon), LTLfWeakNext(self)._delta(i, epsilon)])

    def to_LDLf(self):
        return self._convert().to_LDLf()
        # return LDLfBox(RegExpStar(RegExpPropositional(PLTrue())), LDLfOr([self.f.to_LDLf(), LDLfNot(LDLfEnd())]))

class LTLfRelease(DualBinaryOperatorNNF, BaseConvertibleFormula, LTLfTemporalFormula):
    operator_symbol = "R"
    Dual = LTLfUntil

    def _convert(self):
        return LTLfNot(LTLfUntil([LTLfNot(f) for f in self.formulas]))

    def _delta(self, i:PLInterpretation, epsilon=False):
        if epsilon:
            return PLTrue()
        f1 = self.formulas[0]
        f2 = LTLfRelease(self.formulas[1:]) if len(self.formulas) > 2 else self.formulas[1]
        return PLAnd([
            f2._delta(i, epsilon),
            PLOr([
                f1._delta(i, epsilon),
                LTLfWeakNext(self)._delta(i, epsilon)
            ])
        ])

    def to_LDLf(self):
        return self._convert().to_LDLf()


class LTLfEnd(DeltaConvertibleFormula, BaseConvertibleFormula, LTLfAtomic):
    def __init__(self):
        LTLfAtomic.__init__(self, Symbol("End"))

    def _members(self):
        return (self.s, )

    def _convert(self):
        return LTLfAlways(LTLfFalse()).to_nnf()

    def negate(self):
        return LTLfEventually(LTLfTrue()).to_nnf()

    def __str__(self):
        return "_".join(map(str, self._members()))


LTLfAnd.Dual = LTLfOr
LTLfOr.Dual = LTLfAnd

LTLfNext.Dual = LTLfWeakNext
LTLfWeakNext.Dual = LTLfNext

LTLfEventually.Dual = LTLfAlways
LTLfAlways.Dual = LTLfEventually



from abc import abstractmethod
from functools import lru_cache
from typing import Set


from flloat.base.Alphabet import _Alphabet, Alphabet
from flloat.base.Formula import Formula, CommutativeBinaryOperator, BinaryOperator, AtomicFormula
from flloat.base.Symbol import Symbol
from flloat.base.Symbols import Symbols
from flloat.base.convertible import ImpliesConvertible, EquivalenceConvertible
from flloat.base.nnf import NNF, NotNNF, DualBinaryOperatorNNF, DualCommutativeOperatorNNF
from flloat.base.truths import NotTruth, AndTruth, OrTruth, Truth
from flloat.semantics.pl import PLInterpretation
from flloat.utils import powerset, _powerset, MAX_CACHE_SIZE


class PLTruth(Truth):
    @abstractmethod
    def truth(self, i: PLInterpretation, *args):
        raise NotImplementedError

class PLFormula(Formula, PLTruth, NNF):
    def __init__(self):
        Formula.__init__(self)

        self._all_models = None
        self._minimal_models = None
        self._atoms = None

    def all_models(self, alphabet: _Alphabet) -> Set[PLInterpretation]:
        """Find all the possible interpretations given a set of symbols"""

        all_possible_interpretations = alphabet.powerset().symbols
        all_models = set()
        for i in all_possible_interpretations:
            # compute current Interpretation, considering False
            # all propositional symbols not present in current interpretation
            current_interpretation = PLInterpretation(i)
            if self.truth(current_interpretation):
                all_models.add(current_interpretation)

        self._all_models = all_models
        return all_models

    @lru_cache(maxsize=MAX_CACHE_SIZE)
    def minimal_models(self, alphabet: _Alphabet) -> Set[PLInterpretation]:
        """Find models of min size (i.e. the less number of proposition to True).
        Very trivial (and inefficient) algorithm: BRUTE FORCE on all the possible interpretations."""
        models = self.all_models(alphabet)

        minimal_models = set()
        for m in models:
            min_m = m
            for m1 in models:
                if min_m.true_propositions.issuperset(m1.true_propositions):
                    min_m = m1
            minimal_models.add(min_m)

        return minimal_models

    def __repr__(self):
        return self.__str__()

    def find_atomics(self):
        if self._atoms is None:
            self._atoms = self._find_atomics()
        return self._atoms

    @abstractmethod
    def _find_atomics(self):
        raise NotImplementedError

class PLBinaryOperator(BinaryOperator, PLFormula):
    # def __init__(self, formulas):
    #     PLFormula.__init__(self)
    #     BinaryOperator.__init__(self, formulas)

    def _find_atomics(self):
        res = set()
        for subf in self.formulas:
            try:
                res = res.union(subf.find_atomics())
            except:
                res.add(subf)
        return res

class PLCommBinaryOperator(DualCommutativeOperatorNNF, PLFormula):
    # def __init__(self, formulas):
    #     PLFormula.__init__(self)
    #     DualCommutativeOperatorNNF.__init__(self, formulas)

    def _find_atomics(self):
        res = set()
        for subf in self.formulas:
            try:
                res = res.union(subf.find_atomics())
            except:
                res.add(subf)
        return res


class PLAtomic(AtomicFormula, PLFormula):
    # def __init__(self, s):
    #     PLFormula.__init__(self)
    #     AtomicFormula.__init__(self, s)

    def truth(self, i:PLInterpretation, *args):
        return self.s in i

    def _to_nnf(self):
        return self

    def negate(self):
        return PLNot(self)

    def find_labels(self):
        return {self.s}

    def _find_atomics(self):
        return {self}

class PLTrue(PLAtomic):
    def __init__(self):
        PLAtomic.__init__(self, Symbol(Symbols.TRUE.value))

    def truth(self, *args):
        return True

    def negate(self):
        return PLFalse()

    def find_labels(self):
        return set()

class PLFalse(PLAtomic):
    def __init__(self):
        PLAtomic.__init__(self, Symbol(Symbols.FALSE.value))

    def truth(self, *args):
        return False

    def negate(self):
        return PLTrue()

    def find_labels(self):
        return set()


class PLNot(NotTruth, PLFormula, NotNNF):

    def _find_atomics(self):
        # try:
        #     return self.f.find_atomics()
        # except:
        #     return self.f
        return self.f.find_atomics()


class PLOr(PLCommBinaryOperator, OrTruth):
    pass


class PLAnd(PLCommBinaryOperator, AndTruth):
    pass


class PLImplies(PLBinaryOperator, ImpliesConvertible):
    And = PLAnd
    Or = PLOr
    Not = PLNot


class PLEquivalence(PLCommBinaryOperator, EquivalenceConvertible):
    And = PLAnd
    Or = PLOr
    Not = PLNot

PLOr.Dual = PLAnd
PLAnd.Dual = PLOr



from flloat.base.Symbol import Symbol
from flloat.parser.ldlf import LDLfParser
from flloat.semantics.ldlf import FiniteTrace
from flloat.semantics.pl import PLInterpretation, PLFalseInterpretation
from flloat.syntax.ldlf import LDLfLogicalTrue, LDLfLogicalFalse, LDLfNot, LDLfAnd, LDLfPropositional, \
    RegExpPropositional, LDLfDiamond, LDLfEquivalence, LDLfBox, RegExpStar, LDLfOr, RegExpUnion, RegExpSequence, \
    LDLfEnd, RegExpTest, LDLfLast
from flloat.syntax.pl import PLAnd, PLAtomic, PLNot, PLEquivalence, PLFalse, PLTrue


def test_parser():
    parser = LDLfParser()
    sa, sb = Symbol("A"), Symbol("B")
    a, b = PLAtomic(sa), PLAtomic(sb)

    tt = LDLfLogicalTrue()
    ff = LDLfLogicalFalse()
    true = PLTrue()
    false = PLFalse()
    r_true = RegExpPropositional(true)
    r_false = RegExpPropositional(false)

    assert tt == parser("tt")
    assert ff == parser("ff")
    assert LDLfDiamond(r_true, tt) == parser("<true>tt")
    assert LDLfDiamond(r_false, tt) == parser("<false>tt")
    assert parser("!tt & <!A&B>tt") == LDLfAnd([LDLfNot(tt), LDLfDiamond(RegExpPropositional(PLAnd([PLNot(a),b])), tt)])

    assert parser("[true*](([true]ff) | (<!A>tt) | (<(true)*>(<B>tt)))") ==\
        LDLfBox(RegExpStar(r_true),
            LDLfOr([
                LDLfBox(r_true, ff),
                LDLfDiamond(RegExpPropositional(PLNot(a)), tt),
                LDLfDiamond(RegExpStar(r_true), (LDLfDiamond(RegExpPropositional(b), tt)))
            ])
        )

    assert parser("[A&B&A]ff <-> <A&B&A>tt") == LDLfEquivalence([
        LDLfBox(RegExpPropositional(PLAnd([a, b, a])), ff),
        LDLfDiamond(RegExpPropositional(PLAnd([a,b,a])), tt),
    ])

    assert parser("<A+B>tt")  == LDLfDiamond(RegExpUnion([RegExpPropositional(a), RegExpPropositional(b)]), tt)
    assert parser("<A;B>tt") == LDLfDiamond(RegExpSequence([RegExpPropositional(a), RegExpPropositional(b)]), tt)
    assert parser("<A+(B;A)>end") == LDLfDiamond(
        RegExpUnion([RegExpPropositional(a), RegExpSequence([RegExpPropositional(b), RegExpPropositional(a)])]),
        LDLfEnd()
    )

    assert parser("!(<!(A<->D)+(B;C)*+(!last)?>[(true)*]end)") == LDLfNot(
        LDLfDiamond(
            RegExpUnion([
                RegExpPropositional(PLNot(PLEquivalence([PLAtomic(Symbol("A")), PLAtomic(Symbol("D"))]))),
                RegExpStar(RegExpSequence([
                    RegExpPropositional(PLAtomic(Symbol("B"))),
                    RegExpPropositional(PLAtomic(Symbol("C"))),
                ])),
                RegExpTest(LDLfNot(LDLfLast()))
            ]),
            LDLfBox(
                RegExpStar(RegExpPropositional(PLTrue())),
                LDLfEnd()
            )
        )
    )



def test_truth():
    sa, sb = Symbol("a"), Symbol("b")
    a, b =   PLAtomic(sa), PLAtomic(sb)

    i_ = PLFalseInterpretation()
    i_a = PLInterpretation({sa})
    i_b = PLInterpretation({sb})
    i_ab = PLInterpretation({sa, sb})

    tr_false_a_b_ab = FiniteTrace([
        i_,
        i_a,
        i_b,
        i_ab,
        i_
    ])

    tt = LDLfLogicalTrue()
    ff = LDLfLogicalFalse()

    assert      tt.truth(tr_false_a_b_ab, 0)
    assert not  ff.truth(tr_false_a_b_ab, 0)
    assert not  LDLfNot(tt).truth(tr_false_a_b_ab, 0)
    assert      LDLfNot(ff).truth(tr_false_a_b_ab, 0)
    assert      LDLfAnd([LDLfPropositional(a), LDLfPropositional(b)]).truth(tr_false_a_b_ab, 3)
    assert not  LDLfDiamond(RegExpPropositional(PLAnd([a, b])), tt).truth(tr_false_a_b_ab, 0)

    parser = LDLfParser()
    trace = FiniteTrace.fromStringSets([
        {},
        {"A"},
        {"A"},
        {"A", "B"},
        {}
    ])

    formula = "<true*;A&B>tt"
    parsed_formula = parser(formula)
    assert parsed_formula.truth(trace, 0)

    formula = "[(A+!B)*]<C>tt"
    parsed_formula = parser(formula)
    assert not parsed_formula.truth(trace, 1)

    formula = "<(<!C>tt)?><A>tt"
    parsed_formula = parser(formula)
    assert parsed_formula.truth(trace, 1)

    formula = "<!C+A>tt"
    parsed_formula = parser(formula)
    assert parsed_formula.truth(trace, 1)


def test_nnf():
    parser = LDLfParser()

    assert parser("!tt").to_nnf() == LDLfLogicalFalse()
    assert parser("!!tt").to_nnf() == LDLfLogicalTrue()

    assert parser("!(<!(A&B)>end)").to_nnf() == parser("[!A | !B]<true>tt")

    f = parser("!(<!(A<->D)+(B;C)*+(!last)?>[(true)*]end)")
    assert f.to_nnf() == parser("[(([true]<true>tt)? + ((B ; C))* + ((A | D) & (!(D) | !(A))))]<(true)*><true>tt")
    assert f.to_nnf() == f.to_nnf().to_nnf().to_nnf().to_nnf()

def test_delta():
    parser = LDLfParser()
    sa, sb, sc = Symbol("A"), Symbol("B"), Symbol("C")
    a, b, c = PLAtomic(sa), PLAtomic(sb), PLAtomic(sc)

    i_ = PLFalseInterpretation()
    i_a = PLInterpretation({sa})
    i_b = PLInterpretation({sb})
    i_ab = PLInterpretation({sa, sb})

    true = PLTrue()
    false = PLFalse()
    tt = LDLfLogicalTrue()
    ff = LDLfLogicalFalse()

    assert parser("<A>tt").delta(i_)   == false
    assert parser("<A>tt").delta(i_a)  == tt
    assert parser("<A>tt").delta(i_b)  == false
    assert parser("<A>tt").delta(i_ab) == tt

    assert parser("[B]ff").delta(i_)   == true
    assert parser("[B]ff").delta(i_a)  == true
    assert parser("[B]ff").delta(i_b)  == ff
    assert parser("[B]ff").delta(i_ab) == ff

    # TODO: many other cases!

    f = parser("!(<!(A<->B)+(B;A)*+(!last)?>[(true)*]end)")
    assert f.delta(i_) == f.to_nnf().delta(i_)
    assert f.delta(i_ab) == f.to_nnf().delta(i_ab)

    assert f.delta(i_, epsilon=True) == f.to_nnf().delta(i_, epsilon=True)
    assert f.delta(i_ab, epsilon=True) == f.to_nnf().delta(i_ab, epsilon=True)
    # with epsilon=True, the result is either PLTrue or PLFalse
    assert f.delta(i_, epsilon=True) in [PLTrue(), PLFalse()]


def test_find_labels():
    parser = LDLfParser()

    f = "< (!(A | B | C ))* ; (A | C) ; (!(A | B | C))* ; (B | C) ><true>tt"
    formula = parser(f)
    assert formula.find_labels() == {Symbol(c) for c in "ABC"}

    f = "(<((((<B>tt)?);true)*) ; ((<(A & B)>tt) ?)>tt)"
    formula = parser(f)
    assert formula.find_labels() == {Symbol(c) for c in "AB"}

def test_to_automaton():
    parser = LDLfParser()
    a, b, c = Symbol("A"), Symbol("B"), Symbol("C")
    alphabet_abc = {a, b, c}

    i_ = PLInterpretation(set())
    i_a = PLInterpretation({a})
    i_b = PLInterpretation({b})
    i_ab = PLInterpretation({a, b})

    def _dfa_test(parser, string_formula, alphabet, test_function):
        """temporary function to easily test both the full DFA and the on-the-fly DFA"""
        dfa = parser(string_formula).to_automaton(alphabet, determinize=True, minimize=True)
        test_function(dfa)
        dfa = parser(string_formula).to_automaton(alphabet, on_the_fly=True)
        test_function(dfa)

    ##################################################################################
    f = "<A+!B>tt"
    def test_f(dfa):
        assert not dfa.word_acceptance([])
        assert     dfa.word_acceptance([i_, i_b])
        assert     dfa.word_acceptance([i_a])
        assert not dfa.word_acceptance([i_b])
        assert     dfa.word_acceptance([i_a, i_, i_ab, i_b])
        assert not dfa.word_acceptance([i_b, i_ab])
    _dfa_test(parser, f, alphabet_abc, test_f)
    ##################################################################################

    ##################################################################################
    f = "<true*;B>tt"

    def test_f(dfa):
        assert not dfa.word_acceptance([])
        assert     dfa.word_acceptance([i_, i_b])
        assert not dfa.word_acceptance([i_a])
        assert     dfa.word_acceptance([i_b])
        assert     dfa.word_acceptance([i_a, i_, i_ab, i_b])
        assert     dfa.word_acceptance([i_b, i_ab])

    _dfa_test(parser, f, alphabet_abc, test_f)
    ##################################################################################

    ##################################################################################
    f = "< (!(A | B | C ))* ; (A | C) ; (!(A | B | C))* ; (B | C) ><true>tt"
    def test_f(dfa):
        assert not dfa.word_acceptance([])
        assert not dfa.word_acceptance([i_, i_b])
        assert dfa.word_acceptance([i_a, i_b, i_])
        assert dfa.word_acceptance([i_, i_, i_, i_, i_a, i_, i_ab, i_, i_])
        assert not dfa.word_acceptance([i_b, i_b])
    _dfa_test(parser, f, alphabet_abc, test_f)
    ##################################################################################

    ##################################################################################
    f = "(<((((<B>tt)?);true)*) ; ((<(A & B)>tt) ?)>tt)"
    def test_f(dfa):
        assert not dfa.word_acceptance([])
        assert not dfa.word_acceptance([i_b, i_b, i_b])
        assert dfa.word_acceptance([i_b, i_b, i_ab])
    _dfa_test(parser, f, alphabet_abc, test_f)
    ##################################################################################

    ##################################################################################
    f = "(<true>tt) & ([A]<B>tt)"
    def test_f(dfa):
        assert not dfa.word_acceptance([])
        assert dfa.word_acceptance([i_b])
        assert dfa.word_acceptance([i_])
        assert not dfa.word_acceptance([i_a])
        assert not dfa.word_acceptance([i_ab])
        assert dfa.word_acceptance([i_ab, i_ab])
        assert dfa.word_acceptance([i_a, i_b])
    _dfa_test(parser, f, alphabet_abc, test_f)
    ##################################################################################

    #################################################################################
    f = "[true*](<A>tt -> <true*><B>tt)"
    def test_f(dfa):
        assert dfa.word_acceptance([])
        assert dfa.word_acceptance([i_b])
        assert dfa.word_acceptance([i_])
        assert not dfa.word_acceptance([i_a])
        assert dfa.word_acceptance([i_ab])
        assert dfa.word_acceptance([i_ab, i_ab])
        assert dfa.word_acceptance([i_a, i_b])
        assert not dfa.word_acceptance([i_a, i_a])
    _dfa_test(parser, f, alphabet_abc, test_f)
    #################################################################################

    #################################################################################
    # f = "<true*;A & B;true*;C & B>tt & [((<true>tt)?;true)*](<B -> (A | C)>tt  | [true]ff)"
    #
    # def test_f(dfa):
    #     assert not dfa.word_acceptance([i_])
    #     assert not dfa.word_acceptance([i_b])
    #     assert     dfa.word_acceptance([i_ab])
    #     assert     dfa.word_acceptance([i_])
    #     assert not dfa.word_acceptance([i_a])
    #     assert     dfa.word_acceptance([i_ab])
    #     assert     dfa.word_acceptance([i_ab, i_ab])
    #     assert     dfa.word_acceptance([i_a, i_b])
    #     assert not dfa.word_acceptance([i_a, i_a])
    #
    # _dfa_test(parser, f, alphabet_abc, test_f)
    #################################################################################




from flloat.base.Symbol import Symbol
from flloat.base.Symbols import Symbols
from flloat.base.parsing import Lexer, Parser

from flloat.syntax.ltlf import LTLfNext, LTLfNot, LTLfUntil, LTLfEquivalence, LTLfImplies, LTLfOr, LTLfAnd, \
    LTLfEventually, LTLfAlways, LTLfAtomic, LTLfRelease, LTLfTrue, LTLfFalse, LTLfWeakNext
from flloat.syntax.pl import PLTrue, PLFalse, PLAtomic
from flloat.utils import sym2regexp


class LTLfLexer(Lexer):

    def __init__(self):
        super().__init__()

    reserved = {
        'true':                     'TRUE',
        'false':                    'FALSE',
        Symbols.NEXT.value:         'NEXT',
        Symbols.WEAK_NEXT.value:    'WEAK_NEXT',
        Symbols.UNTIL.value:        'UNTIL',
        Symbols.EVENTUALLY.value:   'EVENTUALLY',
        Symbols.ALWAYS.value:       'ALWAYS',
        Symbols.RELEASE.value:      'RELEASE',
    }

    # List of token names.   This is always required
    tokens = (
        'ATOM',
        'NOT',
        'AND',
        'OR',
        'IMPLIES',
        'EQUIVALENCE',
        'LPAREN',
        'RPAREN',
    ) + tuple(reserved.values())

    # Regular expression rules for simple tokens
    t_NOT               = sym2regexp(Symbols.NOT)
    t_AND               = sym2regexp(Symbols.AND)
    t_OR                = sym2regexp(Symbols.OR)
    t_IMPLIES           = sym2regexp(Symbols.IMPLIES)
    t_EQUIVALENCE       = sym2regexp(Symbols.EQUIVALENCE)
    t_LPAREN            = sym2regexp(Symbols.ROUND_BRACKET_LEFT)
    t_RPAREN            = sym2regexp(Symbols.ROUND_BRACKET_RIGHT)
    t_NEXT              = sym2regexp(Symbols.NEXT)
    t_UNTIL             = sym2regexp(Symbols.UNTIL)
    t_EVENTUALLY        = sym2regexp(Symbols.EVENTUALLY)
    t_ALWAYS            = sym2regexp(Symbols.ALWAYS)
    t_RELEASE           = sym2regexp(Symbols.RELEASE)


    def t_ATOM(self, t):
        r'[a-zA-Z_][a-zA-Z_0-9]*'
        t.type = LTLfLexer.reserved.get(t.value, 'ATOM')  # Check for reserved words
        return t


# Yacc example
class LTLfParser(Parser):

    def __init__(self):
        lexer = LTLfLexer()
        precedence = (
            ('left', 'UNTIL', 'EVENTUALLY', 'ALWAYS', 'RELEASE'),
            ('right', 'NEXT', 'WEAK_NEXT'),
            ('left',  'EQUIVALENCE'),
            ('left',  'IMPLIES'),
            ('left',  'OR'),
            ('left',  'AND'),
            ('right', 'NOT'),

        )
        super().__init__("ltlf", lexer.tokens, lexer, precedence)

    def p_formula(self, p):
        """formula : formula EQUIVALENCE formula
                   | formula IMPLIES formula
                   | formula OR formula
                   | formula AND formula
                   | formula UNTIL formula
                   | formula RELEASE formula
                   | EVENTUALLY formula
                   | ALWAYS formula
                   | NEXT formula
                   | WEAK_NEXT formula
                   | NOT formula
                   | TRUE
                   | FALSE
                   | ATOM"""
        if len(p) == 2:
            if p[1] == Symbols.TRUE.value:
                p[0] = LTLfTrue()
            elif p[1] == Symbols.FALSE.value:
                p[0] = LTLfFalse()
            else:
                p[0] = LTLfAtomic(Symbol(p[1]))
        elif len(p) == 3:
            if p[1] == Symbols.NEXT.value:
                p[0] = LTLfNext(p[2])
            elif p[1] == Symbols.WEAK_NEXT.value:
                p[0] = LTLfWeakNext(p[2])
            elif p[1] == Symbols.EVENTUALLY.value:
                p[0] = LTLfEventually(p[2])
            elif p[1] == Symbols.ALWAYS.value:
                p[0] = LTLfAlways(p[2])
            elif p[1] == Symbols.NOT.value:
                p[0] = LTLfNot(p[2])
        elif len(p) == 4:
            l, o, r = p[1:]
            if o == Symbols.EQUIVALENCE.value:
                p[0] = LTLfEquivalence([l, r])
            elif o == Symbols.IMPLIES.value:
                p[0] = LTLfImplies([l, r])
            elif o == Symbols.OR.value:
                p[0] = LTLfOr([l, r])
            elif o == Symbols.AND.value:
                p[0] = LTLfAnd([l, r])
            elif o == Symbols.UNTIL.value:
                p[0] = LTLfUntil([l, r])
            elif o == Symbols.RELEASE.value:
                p[0] = LTLfRelease([l, r])
            else:
                raise ValueError
        else:
            raise ValueError

    def p_expr_paren(self, p):
        """formula : LPAREN formula RPAREN"""
        p[0] = p[2]


if __name__ == '__main__':
    parser = LTLfParser()
    while True:
        try:
            s = input('calc > ')
        except EOFError:
            break
        if not s: continue
        result = parser(s)
        print(result)




# parsetab.py
# This file is automatically generated. Do not edit.
# pylint: disable=W,C,R
_tabversion = '3.10'

_lr_method = 'LALR'

_lr_signature = 'leftEQUIVALENCErightIMPLIESleftORleftANDrightNOTAND ATOM EQUIVALENCE FALSE IMPLIES LPAREN NOT OR RPAREN TRUEformula : ATOM\n                   | TRUE\n                   | FALSEformula : NOT formulaformula : formula OR formulaformula : formula AND formulaformula : formula IMPLIES formulaformula : formula EQUIVALENCE formulaformula : LPAREN formula RPAREN'
    
_lr_action_items = {'EQUIVALENCE':([1,3,4,6,11,12,13,14,15,16,17,],[8,-2,-1,-3,8,-4,-7,-8,-5,-6,-9,]),'IMPLIES':([1,3,4,6,11,12,13,14,15,16,17,],[7,-2,-1,-3,7,-4,7,7,-5,-6,-9,]),'RPAREN':([3,4,6,11,12,13,14,15,16,17,],[-2,-1,-3,17,-4,-7,-8,-5,-6,-9,]),'OR':([1,3,4,6,11,12,13,14,15,16,17,],[9,-2,-1,-3,9,-4,9,9,-5,-6,-9,]),'LPAREN':([0,2,5,7,8,9,10,],[2,2,2,2,2,2,2,]),'TRUE':([0,2,5,7,8,9,10,],[3,3,3,3,3,3,3,]),'AND':([1,3,4,6,11,12,13,14,15,16,17,],[10,-2,-1,-3,10,-4,10,10,10,-6,-9,]),'ATOM':([0,2,5,7,8,9,10,],[4,4,4,4,4,4,4,]),'NOT':([0,2,5,7,8,9,10,],[5,5,5,5,5,5,5,]),'$end':([1,3,4,6,12,13,14,15,16,17,],[0,-2,-1,-3,-4,-7,-8,-5,-6,-9,]),'FALSE':([0,2,5,7,8,9,10,],[6,6,6,6,6,6,6,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'formula':([0,2,5,7,8,9,10,],[1,11,12,13,14,15,16,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> formula","S'",1,None,None,None),
  ('formula -> ATOM','formula',1,'p_formula_atom','pl.py',61),
  ('formula -> TRUE','formula',1,'p_formula_atom','pl.py',62),
  ('formula -> FALSE','formula',1,'p_formula_atom','pl.py',63),
  ('formula -> NOT formula','formula',2,'p_formula_not','pl.py',72),
  ('formula -> formula OR formula','formula',3,'p_formula_or','pl.py',76),
  ('formula -> formula AND formula','formula',3,'p_formula_and','pl.py',80),
  ('formula -> formula IMPLIES formula','formula',3,'p_formula_implies','pl.py',84),
  ('formula -> formula EQUIVALENCE formula','formula',3,'p_formula_equivalence','pl.py',88),
  ('formula -> LPAREN formula RPAREN','formula',3,'p_formula_expression','pl.py',92),
]



from enum import Enum


class Symbols(Enum):
    NOT = "!"
    AND = "&"
    OR = "|"
    # EXISTS = "∃"
    # FORALL = "Ɐ"
    EQUAL = "="
    IMPLIES = "->"
    EQUIVALENCE = "<->"
    NEXT = "X"
    WEAK_NEXT = "WX"
    UNTIL = "U"
    RELEASE = "R"
    EVENTUALLY = "F"
    ALWAYS = "G"
    PATH_UNION = "+"
    PATH_SEQUENCE = ";"
    PATH_STAR = "*"
    PATH_TEST = "?"
    ROUND_BRACKET_LEFT = "("
    ROUND_BRACKET_RIGHT = ")"
    EVENTUALLY_BRACKET_LEFT = "<"
    EVENTUALLY_BRACKET_RIGHT = ">"
    ALWAYS_BRACKET_LEFT = "["
    ALWAYS_BRACKET_RIGHT = "]"
    # TOP = "⊤"
    # BOTTOM = "⊥"
    LAST = "last"
    END = "end"
    # LOGICAL_TRUE = "⊤⊤"
    # LOGICAL_FALSE = "⊥⊥"
    LOGICAL_TRUE = "tt"
    LOGICAL_FALSE = "ff"
    CARET = "^"
    TRUE="true"
    FALSE = "false"
    LTLf_LAST = "last"


ALL_SYMBOLS = {v.value for v in Symbols}



from flloat.base.Symbol import Symbol
from flloat.base.Symbols import Symbols
from flloat.base.parsing import Lexer, Parser

from flloat.syntax.ldlf import LDLfLogicalTrue, LDLfLogicalFalse, LDLfNot, LDLfOr, LDLfEquivalence, LDLfImplies, \
    LDLfAnd, LDLfDiamond, LDLfBox, RegExpTest, RegExpStar, RegExpUnion, RegExpSequence, RegExpPropositional, LDLfEnd, \
    LDLfLast
from flloat.syntax.pl import PLNot, PLAtomic, PLOr, PLAnd, PLImplies, PLEquivalence, PLTrue, PLFalse
from flloat.utils import sym2regexp


class LDLfLexer(Lexer):

    def __init__(self):
        super().__init__()

    reserved = {
        'true':  'TRUE',
        'false': 'FALSE',
        'tt':    'TT',
        'ff':    'FF',
        'end':   'END',
        'last':  'LAST',
    }

    # List of token names.   This is always required
    tokens = (
        'ATOM',
        'NOT',
        'AND',
        'OR',
        'IMPLIES',
        'EQUIVALENCE',
        'TEST',
        'SEQ',
        'UNION',
        'STAR',
        'LPAREN',
        'RPAREN',
        'BOXLSEPARATOR',
        'BOXRSEPARATOR',
        'DIAMONDLSEPARATOR',
        'DIAMONDRSEPARATOR',
    ) + tuple(reserved.values())

    # Regular expression rules for simple tokens
    t_NOT               = sym2regexp(Symbols.NOT)
    t_AND               = sym2regexp(Symbols.AND)
    t_OR                = sym2regexp(Symbols.OR)
    t_IMPLIES           = sym2regexp(Symbols.IMPLIES)
    t_EQUIVALENCE       = sym2regexp(Symbols.EQUIVALENCE)
    t_TEST              = sym2regexp(Symbols.PATH_TEST)
    t_SEQ               = sym2regexp(Symbols.PATH_SEQUENCE)
    t_UNION             = sym2regexp(Symbols.PATH_UNION)
    t_STAR              = sym2regexp(Symbols.PATH_STAR)
    t_LPAREN            = sym2regexp(Symbols.ROUND_BRACKET_LEFT)
    t_RPAREN            = sym2regexp(Symbols.ROUND_BRACKET_RIGHT)
    t_BOXLSEPARATOR     = sym2regexp(Symbols.ALWAYS_BRACKET_LEFT)
    t_BOXRSEPARATOR     = sym2regexp(Symbols.ALWAYS_BRACKET_RIGHT)
    t_DIAMONDLSEPARATOR = sym2regexp(Symbols.EVENTUALLY_BRACKET_LEFT)
    t_DIAMONDRSEPARATOR = sym2regexp(Symbols.EVENTUALLY_BRACKET_RIGHT)

    def t_ATOM(self, t):
        r'[a-zA-Z_][a-zA-Z_0-9]*'
        t.type = LDLfLexer.reserved.get(t.value, 'ATOM')  # Check for reserved words
        return t


# Yacc example
class LDLfParser(Parser):

    def __init__(self):
        lexer = LDLfLexer()
        precedence = (
            ('left',  'EQUIVALENCE'),
            ('left',  'IMPLIES'),
            ('left',  'UNION'),
            ('left',  'SEQ'),
            ('left',  'STAR'),
            ('left',  'TEST'),
            ('left',  'OR'),
            ('left',  'AND'),
            ('right', 'DIAMONDLSEPARATOR', 'BOXLSEPARATOR'),
            ('left',  'DIAMONDRSEPARATOR', 'BOXRSEPARATOR'),
            ('right', 'NOT'),

        )
        super().__init__("ldlf", lexer.tokens, lexer, precedence)

        # self.pl_parser = PLParser()

    def p_temp_formula(self, p):
        """temp_formula : temp_formula EQUIVALENCE temp_formula
                        | temp_formula IMPLIES temp_formula
                        | temp_formula OR temp_formula
                        | temp_formula AND temp_formula
                        | BOXLSEPARATOR path BOXRSEPARATOR temp_formula
                        | DIAMONDLSEPARATOR path DIAMONDRSEPARATOR temp_formula
                        | NOT temp_formula
                        | TT
                        | FF
                        | END
                        | LAST"""
        if len(p) == 2:
            if p[1] == Symbols.LOGICAL_TRUE.value:
                p[0] = LDLfLogicalTrue()
            elif p[1] == Symbols.LOGICAL_FALSE.value:
                p[0] = LDLfLogicalFalse()
            elif p[1] == Symbols.END.value:
                p[0] = LDLfEnd()
            elif p[1] == Symbols.LAST.value:
                p[0] = LDLfLast()
            else:
                p[0] = LDLfDiamond(RegExpPropositional(p[1]), LDLfLogicalTrue())
        elif len(p) == 3:
            p[0] = LDLfNot(p[2])
        elif len(p) == 4:
            l, o, r = p[1:]
            if o == Symbols.EQUIVALENCE.value:
                p[0] = LDLfEquivalence([l, r])
            elif o == Symbols.IMPLIES.value:
                p[0] = LDLfImplies([l, r])
            elif o == Symbols.OR.value:
                p[0] = LDLfOr([l, r])
            elif o == Symbols.AND.value:
                p[0] = LDLfAnd([l, r])
            else:
                raise ValueError
        elif len(p) == 5:
            if p[1] == Symbols.ALWAYS_BRACKET_LEFT.value:
                p[0] = LDLfBox(p[2], p[4])
            elif p[1] == Symbols.EVENTUALLY_BRACKET_LEFT.value:
                p[0] = LDLfDiamond(p[2], p[4])
            else:
                raise ValueError
        else:
            raise ValueError

    # def p_formula_propositional(self, p):
    #     'formula : propositional'
    #     p[0] = LDLfDiamond(RegExpPropositional(p[1]), LDLfLogicalTrue())

    def p_path(self, p):
        """path : path UNION path
                | path SEQ path
                | path STAR
                | temp_formula TEST
                | propositional"""

        if len(p)==2:
            p[0] = RegExpPropositional(p[1])
        elif len(p)==3:
            if p[2]==Symbols.PATH_TEST.value:
                p[0] = RegExpTest(p[1])
            elif p[2] == Symbols.PATH_STAR.value:
                p[0] = RegExpStar(p[1])
            else:
                raise ValueError
        elif len(p)==4:
            if p[2]==Symbols.PATH_UNION.value:
                p[0] = RegExpUnion([p[1], p[3]])
            elif p[2] == Symbols.PATH_SEQUENCE.value:
                p[0] = RegExpSequence([p[1], p[3]])
            else:
                raise ValueError
        else:
            raise ValueError


    def p_propositional(self, p):
        """propositional : propositional EQUIVALENCE propositional
                         | propositional IMPLIES propositional
                         | propositional OR propositional
                         | propositional AND propositional
                         | NOT propositional
                         | FALSE
                         | TRUE
                         | ATOM"""
        if len(p)==4:
            if p[2] == Symbols.EQUIVALENCE.value:
                p[0] = PLEquivalence([p[1], p[3]])
            elif p[2] == Symbols.IMPLIES.value:
                p[0] = PLImplies([p[1], p[3]])
            elif p[2] == Symbols.OR.value:
                p[0] = PLOr([p[1], p[3]])
            elif p[2] == Symbols.AND.value:
                p[0] = PLAnd([p[1], p[3]])
            else:
                raise ValueError
            # else:
            #     p[0] = p[2]
        elif len(p)==3:
            p[0] = PLNot(p[2])
        elif len(p)==2:
            if p[1]==Symbols.TRUE.value:
                p[0] = PLTrue()
            elif p[1]==Symbols.FALSE.value:
                p[0] = PLFalse()
            else:
                p[0] = PLAtomic(Symbol(p[1]))
        else:
            raise ValueError

    def p_expr_paren(self, p):
        """temp_formula : LPAREN temp_formula RPAREN
        path            : LPAREN path RPAREN
        propositional   : LPAREN propositional RPAREN
        """
        p[0] = p[2]




if __name__ == '__main__':
    parser = LDLfParser()
    while True:
        try:
            s = input('calc > ')
        except EOFError:
            break
        if not s: continue
        result = parser(s)
        print(result)



from typing import Set, List

from pythomata.base.Alphabet import Alphabet as PythomataAlphabet

from pythomata.base.DFA import DFA
from pythomata.base.NFA import NFA
from pythomata.base.Simulator import Simulator
from pythomata.base.utils import MacroState

from flloat.base.Alphabet import Alphabet
from flloat.base.Formula import Formula
from flloat.base.Symbol import Symbol
from flloat.semantics.pl import PLInterpretation, PLTrueInterpretation, PLFalseInterpretation
from flloat.syntax.pl import PLFormula, PLAtomic, PLTrue, PLFalse, PLNot, PLBinaryOperator, PLAnd, PLOr, PLImplies, \
    PLEquivalence
from flloat.utils import powerset

def find_atomics(formula: Formula) -> Set[PLAtomic]:
    """Finds all the atomic formulas"""
    f = formula
    res = set()
    if isinstance(formula, PLFormula):
        res = formula.find_atomics()
    # elif isinstance(f, PLNot):
    #     res = res.union(find_atomics(f.f))
    # elif isinstance(f, PLBinaryOperator):
    #     for subf in f.formulas:
    #         res = res.union(find_atomics(subf))
    else:
        res.add(f)
    return res

def _transform_delta(f:Formula, formula2AtomicFormula):
    """From a Propositional Formula to a Propositional Formula
    with non-propositional subformulas replaced with a "freezed" atomic formula."""
    t = type(f)
    if t == PLNot:
        return PLNot(_transform_delta(f, formula2AtomicFormula))
    # elif isinstance(f, PLBinaryOperator): #PLAnd, PLOr, PLImplies, PLEquivalence
    elif t == PLAnd or t == PLOr or t == PLImplies or t == PLEquivalence:
        return t([_transform_delta(subf, formula2AtomicFormula) for subf in f.formulas])
    elif t == PLTrue or t == PLFalse:
        return f
    else:
        return formula2AtomicFormula[f]


def to_automaton_(f, labels:Set[Symbol]=None):
    """
    DEPRECATED
    From a LDLfFormula, build the automaton.
    :param f:               a LDLfFormula;
    :param labels:          a set of Symbol, the fluents of our domain. If None, retrieve them from the formula;
    :param determinize:     True if you need to determinize the NFA, obtaining a DFA;
    :param minimize:        True if you need to minimize the DFA (if determinize is False this flag has no effect.)
    :return:                a NFA or a DFA which accepts the same traces that makes the formula True.
    """

    nnf = f.to_nnf()

    if labels is None:
        # if the labels of the formula are not specified in input,
        # retrieve them from the formula
        labels = nnf.find_labels()

    # the alphabet is the powerset of the set of fluents
    alphabet = powerset(labels)
    initial_state = MacroState({nnf})
    final_states = {MacroState()}
    delta = set()

    d = f.delta(PLFalseInterpretation(), epsilon=True)
    if d.truth(d):
        final_states.add(initial_state)

    states = {MacroState(), initial_state}

    states_changed, delta_changed = True, True

    while states_changed or delta_changed:

        states_changed, delta_changed = False, False
        for actions_set in alphabet:
            states_list = list(states)
            for q in states_list:

                # delta function applied to every formula in the macro state Q
                delta_formulas = [f.delta(actions_set) for f in q]

                # find the list of atoms, which are "true" atoms (i.e. propositional atoms) or LDLf formulas
                atomics = [s for subf in delta_formulas for s in find_atomics(subf)]

                # "freeze" the found atoms as symbols and build a mapping from symbols to formulas
                symbol2formula = {Symbol(str(f)): f for f in atomics if f != PLTrue() and f != PLFalse()}

                # build a map from formula to a "freezed" propositional Atomic Formula
                formula2atomic_formulas = {
                    f: PLAtomic(Symbol(str(f)))
                    if f != PLTrue() and f != PLFalse()# and not isinstance(f, PLAtomic)
                    else f for f in atomics
                }

                # the final list of Propositional Atomic Formulas, one for each formula in the original macro state Q
                transformed_delta_formulas = [_transform_delta(f, formula2atomic_formulas) for f in delta_formulas]

                # the empty conjunction stands for true
                if len(transformed_delta_formulas) == 0:
                    conjunctions = PLTrue()
                elif len(transformed_delta_formulas) == 1:
                    conjunctions = transformed_delta_formulas[0]
                else:
                    conjunctions = PLAnd(transformed_delta_formulas)

                # the model in this case is the smallest set of symbols s.t. the conjunction of "freezed" atomic formula
                # is true.
                models = frozenset(conjunctions.minimal_models(Alphabet(symbol2formula)))

                if len(models) == 0:
                    continue
                for min_model in models:
                    q_prime = MacroState(
                        {symbol2formula[s] for s in min_model.true_propositions})

                    len_before = len(states)
                    states.add(q_prime)
                    if len(states) == len_before + 1:
                        states_list.append(q_prime)
                        states_changed = True

                    len_before = len(delta)
                    delta.add((q, actions_set, q_prime))
                    if len(delta) == len_before + 1:
                        delta_changed = True

                    # check if q_prime should be added as final state
                    if len(q_prime) == 0:
                        final_states.add(q_prime)
                    else:
                        subf_deltas = [subf.delta(PLFalseInterpretation(), epsilon=True) for subf in q_prime]
                        if len(subf_deltas)==1:
                            q_prime_delta_conjunction = subf_deltas[0]
                        else:
                            q_prime_delta_conjunction = PLAnd(subf_deltas)

                        if q_prime_delta_conjunction.truth(PLFalseInterpretation()):
                            final_states.add(q_prime)


    alphabet = PythomataAlphabet({PLInterpretation(set(sym)) for sym in alphabet})
    delta = frozenset((i, PLInterpretation(set(a)), o) for i, a, o in delta)


    nfa = NFA.fromTransitions(
        alphabet=alphabet,
        states=frozenset(states),
        initial_state=initial_state,
        accepting_states=frozenset(final_states),
        transitions=delta
    )

    return nfa


class DFAOTF(Simulator):
    """DFA on the fly"""

    def __init__(self, f):
        self.f = f.to_nnf()
        self.reset()

    def reset(self):
        self.cur_state = frozenset([frozenset([self.f])])

    def word_acceptance(self, action_set_list:List[PLInterpretation]):
        self.reset()
        for a in action_set_list:
            self.make_transition(a)
        return self.is_true()

    def is_true(self):
        return self._is_true(self.cur_state)

    @staticmethod
    def _is_true(Q):
        if frozenset() in Q:
            return True
        conj = [PLAnd([subf.delta(None, epsilon=True) for subf in q]) if len(q) >= 2 else
                next(iter(q)).delta(None, epsilon=True) if len(q) == 1 else
                PLFalse() for q in Q]
        if len(conj) == 0:
            return False
        else:
            conj = PLOr(conj) if len(conj) >= 2 else conj[0]
        return conj.truth(None)

    def get_current_state(self):
        return self.cur_state

    def make_transition(self, i:PLInterpretation):
        self.cur_state = self._make_transition(self.cur_state, i)

    @staticmethod
    def _make_transition(Q, i: PLInterpretation):
        actions_set = i.true_propositions
        new_macrostate = set()

        for q in Q:
            # delta function applied to every formula in the macro state Q
            delta_formulas = [f.delta(actions_set) for f in q]

            # find the list of atoms, which are "true" atoms (i.e. propositional atoms) or LDLf formulas
            atomics = [s for subf in delta_formulas for s in find_atomics(subf)]

            atom2id = {v: k for k, v in enumerate(atomics)}
            id2atom = {v: k for k, v in atom2id.items()}

            # "freeze" the found atoms as symbols and build a mapping from symbols to formulas
            symbol2formula = {atom2id[f] for f in atomics if f != PLTrue() and f != PLFalse()}

            # build a map from formula to a "freezed" propositional Atomic Formula
            formula2atomic_formulas = {
                f: PLAtomic(atom2id[f])
                if f != PLTrue() and f != PLFalse()  # and not isinstance(f, PLAtomic)
                else f for f in atomics
            }

            # the final list of Propositional Atomic Formulas, one for each formula in the original macro state Q
            transformed_delta_formulas = [_transform_delta(f, formula2atomic_formulas) for f in delta_formulas]

            # the empy conjunction stands for true
            if len(transformed_delta_formulas) == 0:
                conjunctions = PLTrue()
            elif len(transformed_delta_formulas) == 1:
                conjunctions = transformed_delta_formulas[0]
            else:
                conjunctions = PLAnd(transformed_delta_formulas)

            # the model in this case is the smallest set of symbols s.t. the conjunction of "freezed" atomic formula
            # is true.
            models = frozenset(conjunctions.minimal_models(Alphabet(symbol2formula)))

            for min_model in models:
                q_prime = frozenset({id2atom[s] for s in min_model.true_propositions})

                new_macrostate.add(q_prime)

        return frozenset(new_macrostate)


def to_automaton(f, labels:Set[Symbol]=None, minimize=True):

    initial_state = frozenset({frozenset({f.to_nnf()})})
    states = {initial_state}
    final_states = set()
    transition_function = {}

    # the alphabet is the powerset of the set of fluents
    alphabet = powerset(labels)

    if DFAOTF._is_true(initial_state):
        final_states.add(initial_state)

    visited = set()
    to_be_visited = {initial_state}

    while len(to_be_visited)!=0:

        for q in list(to_be_visited):
            to_be_visited.remove(q)
            for actions_set in alphabet:
                new_state = DFAOTF._make_transition(q, PLInterpretation(actions_set))
                if not new_state in states:
                    states.add(new_state)
                    to_be_visited.add(new_state)

                transition_function.setdefault(q, {})[PLInterpretation(actions_set)] = new_state

                if new_state not in visited:
                    visited.add(new_state)
                    if DFAOTF._is_true(new_state): final_states.add(new_state)


    new_alphabet = PythomataAlphabet({PLInterpretation(set(sym)) for sym in alphabet})
    dfa = DFA(new_alphabet, frozenset(states), initial_state, frozenset(final_states), transition_function)

    if minimize:
        dfa = dfa.minimize().trim()

    return dfa




"""Misc tests"""
import os


def test_ldlf_example_readme():
    from flloat.parser.ldlf import LDLfParser
    from flloat.base.Symbol import Symbol
    from flloat.semantics.ldlf import FiniteTrace

    parser = LDLfParser()
    formula = "<true*; A & B>tt"
    parsed_formula = parser(formula)

    assert str(parsed_formula) == "<((true)* ; (B & A))>(tt)" or str(parsed_formula) == "<((true)* ; (A & B))>(tt)"
    assert parsed_formula.find_labels() == {Symbol(c) for c in "AB"}

    from flloat.semantics.ldlf import FiniteTrace

    t1 = FiniteTrace.fromStringSets([
        {},
        {"A"},
        {"A"},
        {"A", "B"},
        {}
    ])
    assert parsed_formula.truth(t1, 0)

    t2 = FiniteTrace.fromStringSets([
        {},
        {"A"},
        {"B"}
    ])
    assert not parsed_formula.truth(t2, 0)

    dfa = parsed_formula.to_automaton(determinize=True)
    assert     dfa.word_acceptance(t1.trace)
    assert not dfa.word_acceptance(t2.trace)


def test_ltlf_example_readme():
    from flloat.parser.ltlf import LTLfParser
    from flloat.base.Symbol import Symbol
    from flloat.semantics.ldlf import FiniteTrace

    parser = LTLfParser()
    formula = "F (A & !B)"
    parsed_formula = parser(formula)

    t1 = FiniteTrace.fromStringSets([
        {},
        {"A"},
        {"A"},
        {"A", "B"}
    ])
    assert parsed_formula.truth(t1, 0)

    t2 = FiniteTrace.fromStringSets([
        {},
        {"A", "B"},
        {"B"}
    ])
    assert not parsed_formula.truth(t2, 0)

    dfa = parsed_formula.to_automaton(determinize=True)
    assert dfa.word_acceptance(t1.trace)
    assert not dfa.word_acceptance(t2.trace)


def test_hash_consistency_after_pickling():
    from flloat.parser.ltlf import LTLfParser
    import pickle

    parser = LTLfParser()
    formula = "F (A & !B)"
    old_obj = parser(formula)

    h = hash(old_obj)
    pickle.dump(old_obj, open("temp", "wb"))
    new_obj = pickle.load(open("temp", "rb"))

    assert new_obj._hash is None
    assert h==hash(new_obj)

    os.remove("temp")



from abc import ABC

import sys
from ply import lex
import ply.yacc as yacc

import os


class Lexer(ABC):

    @property
    def tokens(self):
        raise NotImplementedError

    # Define a rule so we can track line numbers
    def t_newline(self, t):
        r'\n+'
        t.lexer.lineno += len(t.value)

    # A string containing ignored characters (spaces and tabs)
    t_ignore = ' \t'

    # Error handling rule
    def t_error(self, t):
        print("Illegal character '%s'" % t.value[0])
        t.lexer.skip(1)

    # Build the lexer
    def build(self, **kwargs):
        self.lexer = lex.lex(module=self, **kwargs)


class Parser(ABC):

    def __init__(self, name, tokens, lexer, precedence=None):
        self.name = name
        self.tokens = tokens
        if precedence:
            self.precedence = precedence
        self.lexer = lexer
        self.lexer.build()

        # Build the parser
        # outputdir = "flloat/base/tables/%s" % name
        self.parser = yacc.yacc(module=self)#, outputdir=outputdir)

    def __call__(self, s, **kwargs):
        return self.parser.parse(s, lexer=self.lexer.lexer)

    # Error rule for syntax errors
    def p_error(self, p):
        raise ValueError("Syntax error in input! %s" %str(p))



from abc import abstractmethod, ABC
from functools import lru_cache
from typing import Set

from flloat.base.Formula import Formula, CommutativeBinaryOperator, UnaryOperator, BinaryOperator, OperatorChilds, \
    AtomicFormula
from flloat.base.Symbol import Symbol
from flloat.base.Symbols import Symbols
from flloat.base.convertible import DeltaConvertibleFormula, ImpliesDeltaConvertible, EquivalenceDeltaConvertible, \
    ConvertibleFormula
from flloat.base.misc import Delta
from flloat.base.nnf import NNF, NotNNF, DualBinaryOperatorNNF, DualNNF
from flloat.base.truths import NotTruth, AndTruth, OrTruth, Truth
from flloat.utils import MAX_CACHE_SIZE
from flloat.flloat import to_automaton, DFAOTF, to_automaton_
from flloat.semantics.ldlf import FiniteTrace, FiniteTraceTruth
from flloat.semantics.pl import PLInterpretation, PLFalseInterpretation
from flloat.syntax.pl import PLFormula, PLTrue, PLFalse, PLAnd, PLOr


class RegExpTruth(Truth):
    @abstractmethod
    def truth(self, tr: FiniteTrace, start: int=0, end: int=0):
        raise NotImplementedError

class LDLfFormula(Formula, FiniteTraceTruth, NNF, Delta):

    @lru_cache(maxsize=MAX_CACHE_SIZE)
    def delta(self, i: PLInterpretation, epsilon=False):

        f = self.to_nnf()
        d = f._delta(i, epsilon)
        if epsilon:
            # By definition, if epsilon=True, then the result must be either PLTrue or PLFalse
            # Now, the output is a Propositional Formula with only PLTrue or PLFalse as atomics
            # Hence, we just evaluate the formula with a dummy PLInterpretation
            d = PLTrue() if d.truth(PLFalseInterpretation()) else PLFalse()
        return d

    @abstractmethod
    def _delta(self, i: PLInterpretation, epsilon=False):
        """apply delta function, assuming that 'self' is a LDLf formula in Negative Normal Form"""
        raise NotImplementedError

    def __repr__(self):
        return self.__str__()

    def to_automaton(self, labels:Set[Symbol]=None, on_the_fly=False, determinize=False, minimize=True):
        if labels is None:
            labels = self.find_labels()
        if on_the_fly:
            return DFAOTF(self)
        elif determinize:
            return to_automaton(self, labels, minimize)
        else:
            return to_automaton_(self, labels)


class LDLfCommBinaryOperator(CommutativeBinaryOperator, LDLfFormula):
    pass


class DeltaRegExp(ABC):
    @abstractmethod
    def deltaDiamond(self, f:LDLfFormula, i: PLInterpretation, epsilon=False):
        raise NotImplementedError

    @abstractmethod
    def deltaBox(self, f:LDLfFormula, i: PLInterpretation, epsilon=False):
        raise NotImplementedError

class RegExpFormula(Formula, RegExpTruth, NNF, DeltaRegExp):

    # this should be never called. Just for override the inherited abstract method.
    def negate(self):
        raise NotImplementedError


class LDLfTemporalFormula(LDLfFormula):
    @property
    def temporal_brackets(self)->str:
        raise NotImplementedError

    def __init__(self, r:RegExpFormula, f:LDLfFormula):
        super().__init__()
        self.r = r
        self.f = f

    def _members(self):
        return (self.temporal_brackets, self.r, self.f)

    def __str__(self):
        return self.temporal_brackets[0] + str(self.r) + self.temporal_brackets[1] + "(" + str(self.f) + ")"

    def find_labels(self):
        return self.f.find_labels().union(self.r.find_labels())


class LDLfTemporalFormulaNNF(LDLfTemporalFormula, DualNNF):
    def _to_nnf(self):
        return type(self)(self.r.to_nnf(), self.f.to_nnf())

    def negate(self):
        return self.Dual(self.r, LDLfNot(self.f))


class LDLfAtomic(AtomicFormula, LDLfFormula):

    def __str__(self):
        return AtomicFormula.__str__(self)

    def find_labels(self):
        return set()


class LDLfLogicalTrue(LDLfAtomic):
    def __init__(self):
        super().__init__(Symbol(Symbols.LOGICAL_TRUE.value))

    def truth(self, *args):
        return True

    def _to_nnf(self):
        return self

    def negate(self):
        return LDLfLogicalFalse()

    def _delta(self, i:PLInterpretation, epsilon=False):
        return PLTrue()


class LDLfLogicalFalse(LDLfAtomic):
    def __init__(self):
        super().__init__(Symbol(Symbols.LOGICAL_FALSE.value))

    def truth(self, *args):
        return False

    def _to_nnf(self):
        return self

    def negate(self):
        return LDLfLogicalTrue()

    def _delta(self, i:PLInterpretation, epsilon=False):
        return PLFalse()


class LDLfNot(NotTruth, LDLfFormula, NotNNF):

    # def __init__(self, f:LDLfFormula):
    #     LDLfFormula.__init__(self)
    #     NotTruth.__init__(self, f)

    def _to_nnf(self):
        neg = self.f.negate()
        return neg.to_nnf()

    def negate(self):
        return self.f

    def _delta(self, i: PLInterpretation, epsilon=False):
        # should never called, since it is called from NNF formulas
        raise Exception


class LDLfAnd(LDLfCommBinaryOperator, AndTruth, DualBinaryOperatorNNF):

    def _delta(self, i:PLInterpretation, epsilon=False):
        return PLAnd([f._delta(i, epsilon) for f in self.formulas])


class LDLfOr(LDLfCommBinaryOperator, OrTruth, DualBinaryOperatorNNF):

    def _delta(self, i:PLInterpretation, epsilon=False):
        return PLOr([f._delta(i, epsilon) for f in self.formulas])


class LDLfImplies(ImpliesDeltaConvertible, LDLfFormula):
    And = LDLfAnd
    Or =  LDLfOr
    Not = LDLfNot

class LDLfEquivalence(EquivalenceDeltaConvertible, LDLfCommBinaryOperator):
    And = LDLfAnd
    Or = LDLfOr
    Not = LDLfNot

class LDLfDiamond(LDLfTemporalFormulaNNF, FiniteTraceTruth):
    temporal_brackets = "<>"

    def truth(self, i: FiniteTrace, pos: int=0):
        return any(self.r.truth(i, pos, j) and self.f.truth(i, j) for j in range(pos, i.last()+1))
        # last + 1 in order to include the last step

    def _delta(self, i:PLInterpretation, epsilon=False):
        return self.r.deltaDiamond(self.f, i, epsilon)


class LDLfBox(ConvertibleFormula, LDLfTemporalFormulaNNF):
    temporal_brackets = "[]"

    def _convert(self):
        return LDLfNot(LDLfDiamond(self.r, LDLfNot(self.f)))

    def truth(self, i: FiniteTrace, pos: int=0):
        return self._convert().truth(i, pos)

    def _delta(self, i:PLInterpretation, epsilon=False):
        d = self.r.deltaBox(self.f, i, epsilon)
        return d


class RegExpPropositional(RegExpFormula, PLFormula):
    def __init__(self, pl_formula:PLFormula):
        RegExpFormula.__init__(self)
        self.pl_formula = pl_formula

    def truth(self, tr: FiniteTrace, start: int=0, end: int=0):
        return end == start + 1 \
                and end <= tr.last() \
                and self.pl_formula.truth(tr.get(start))

    def _members(self):
        return (RegExpPropositional, self.pl_formula)

    def __str__(self):
        return str(self.pl_formula)

    def _to_nnf(self):
        return RegExpPropositional(self.pl_formula.to_nnf())

    def negate(self):
        return RegExpPropositional(self.pl_formula.negate())

    def deltaDiamond(self, f:LDLfFormula, i: PLInterpretation, epsilon=False):
        if epsilon:
            return PLFalse()
        if self.pl_formula.truth(i):
            # return PLAtomic(Symbol(str(_expand(f))))
            return _expand(f)
        else:
            return PLFalse()

    def deltaBox(self, f:LDLfFormula, i: PLInterpretation, epsilon=False):
        if epsilon:
            return PLTrue()
        if self.pl_formula.truth(i):
            # return PLAtomic(Symbol(str(_expand(f))))
            return _expand(f)
        else:
            return PLTrue()

    def find_labels(self):
        return self.pl_formula.find_labels()

    def _find_atomics(self):
        return self.pl_formula.find_atomics()

class RegExpTest(UnaryOperator, RegExpFormula):
    operator_symbol = Symbols.PATH_TEST.value

    # def __init__(self, f:LDLfFormula):
    #     RegExpFormula.__init__(self)
    #     UnaryOperator.__init__(self, f)

    def truth(self, tr: FiniteTrace, start: int=0, end: int=0):
        return start == end and self.f.truth(tr, start)

    def __str__(self):
        s = super().__str__()
        s = s[1:] + s[0]
        return s

    def _to_nnf(self):
        return RegExpTest(self.f.to_nnf())

    def deltaDiamond(self, f:LDLfFormula, i: PLInterpretation, epsilon=False):
        return PLAnd([self.f._delta(i, epsilon), f._delta(i, epsilon)])

    def deltaBox(self, f:LDLfFormula, i: PLInterpretation, epsilon=False):
        # ff = LDLfNot(self.f).to_nnf()
        # dff = ff._delta(i, epsilon)
        # fun = f._delta(i, epsilon)
        return PLOr([LDLfNot(self.f).to_nnf()._delta(i, epsilon), f._delta(i, epsilon)])

    def find_labels(self):
        return self.f.find_labels()

class RegExpUnion(CommutativeBinaryOperator, RegExpFormula):
    operator_symbol = Symbols.PATH_UNION.value
    def __init__(self, formulas):
        RegExpFormula.__init__(self)
        CommutativeBinaryOperator.__init__(self, formulas)

    def truth(self, tr: FiniteTrace, start: int=0, end: int=0):
        return any(f.truth(tr, start, end) for f in self.formulas_set)

    def _to_nnf(self):
        return RegExpUnion([r.to_nnf() for r in self.formulas_set])

    def deltaDiamond(self, f:LDLfFormula, i: PLInterpretation, epsilon=False):
        return PLOr([LDLfDiamond(r, f)._delta(i, epsilon) for r in self.formulas_set])

    def deltaBox(self, f:LDLfFormula, i: PLInterpretation, epsilon=False):
        return PLAnd([LDLfBox(r, f)._delta(i, epsilon) for r in self.formulas_set])

class RegExpSequence(BinaryOperator, RegExpFormula):
    operator_symbol = Symbols.PATH_SEQUENCE.value

    def __init__(self, formulas: OperatorChilds):
        RegExpFormula.__init__(self)
        BinaryOperator.__init__(self, formulas)

    def truth(self, tr: FiniteTrace, start: int=0, end: int=0):
        f1 = self.formulas[0]
        if len(self.formulas) == 2:
            f2 = self.formulas[1]
        else:
            f2 = RegExpSequence(self.formulas[1:])

        return any(f1.truth(tr, start, k) and f2.truth(tr, k, end) for k in range(start, end + 1))


    def _to_nnf(self):
        return RegExpSequence([r.to_nnf() for r in self.formulas])

    def deltaDiamond(self, f:LDLfFormula, i: PLInterpretation, epsilon=False):
        res = LDLfDiamond(self.formulas[-1], f)
        for r in reversed(self.formulas[:-1]):
            res = LDLfDiamond(r, res)
        return res._delta(i, epsilon)

    def deltaBox(self, f:LDLfFormula, i: PLInterpretation, epsilon=False):
        res = LDLfBox(self.formulas[-1], f)
        for r in reversed(self.formulas[:-1]):
            res = LDLfBox(r, res)
        return res._delta(i, epsilon)


class RegExpStar(UnaryOperator, RegExpFormula):
    operator_symbol = Symbols.PATH_STAR.value

    # def __init__(self, f):
    #     UnaryOperator.__init__(self, f)
    #     RegExpFormula.__init__(self)

    def truth(self, tr: FiniteTrace, start: int=0, end: int=0):
        return start == end \
            or any(self.f.truth(tr, start, k) and self.truth(tr, k, end)
                   for k in range(start, end + 1))

    def __str__(self):
        s = super().__str__()
        s = s[1:] + s[0]
        return s

    def _to_nnf(self):
        return RegExpStar(self.f.to_nnf())

    def deltaDiamond(self, f:LDLfFormula, i: PLInterpretation, epsilon=False):
        return PLOr([f._delta(i, epsilon), LDLfDiamond(self.f, F(LDLfDiamond(self, f)))._delta(i, epsilon)])

    def deltaBox(self, f:LDLfFormula, i: PLInterpretation, epsilon=False):
        # subf = LDLfBox(self.f, T(LDLfBox(self, f)))
        # k = subf._delta(i, epsilon)
        # l = [f._delta(i, epsilon), subf]
        # ff = PLAnd(l)
        return PLAnd([f._delta(i, epsilon), LDLfBox(self.f, T(LDLfBox(self, f)))._delta(i, epsilon)])


class LDLfPropositional(DeltaConvertibleFormula, LDLfFormula):
    def __init__(self, pl_formula:PLFormula):
        super().__init__()
        self.pl_formula = pl_formula

    def _convert(self):
        return LDLfDiamond(RegExpPropositional(self.pl_formula), LDLfLogicalTrue())

    def _members(self):
        return (LDLfPropositional, self.pl_formula)

    def find_labels(self):
        return self.pl_formula.find_labels()

    def __str__(self):
        return str(self._convert())


class LDLfEnd(DeltaConvertibleFormula, LDLfAtomic):
    def __init__(self):
        super().__init__(Symbol(Symbols.END.value))

    def _convert(self):
        return LDLfBox(RegExpPropositional(PLTrue()), LDLfLogicalFalse())


class LDLfLast(DeltaConvertibleFormula, LDLfAtomic):
    def __init__(self):
        super().__init__(Symbol(Symbols.LAST.value))

    def _convert(self):
        return LDLfDiamond(RegExpPropositional(PLTrue()), LDLfEnd()._convert())


class F(Formula, Delta, NNF):
    def __init__(self, f: Formula):
        super().__init__()
        self.f = f

    def _members(self):
        return ("F", self.f)

    def __str__(self):
        return "_".join(map(str, self._members()))

    def delta(self, i:PLInterpretation, epsilon=False):
        return self._delta(i, epsilon)

    def _delta(self,i:PLInterpretation, epsilon=False):
        return PLFalse()

    def _to_nnf(self):
        return self

    def negate(self):
        return T(self.f)

    def find_labels(self):
        return super().find_labels()


class T(Formula, Delta, NNF):
    def __init__(self, f: Formula):
        super().__init__()
        self.f = f

    def _members(self):
        return ("T", self.f)

    def __str__(self):
        return "_".join(map(str, self._members()))

    def delta(self, i: PLInterpretation, epsilon=False):
        return self._delta(i, epsilon)

    def _delta(self, i:PLInterpretation, epsilon=False):
        return PLTrue()

    def _to_nnf(self):
        return self

    def negate(self):
        return F(self.f)

    def find_labels(self):
        return super().find_labels()

def _expand(f:Formula):
    if isinstance(f, F) or isinstance(f, T):
        return _expand(f.f)
    # elif isinstance(f, LDLfLogicalTrue):
    #     return PLTrue()
    # elif isinstance(f, LDLfLogicalFalse):
    #     return PLFalse()
    elif isinstance(f, LDLfDiamond) or isinstance(f, LDLfBox):
        return type(f)(f.r, _expand(f.f))
    elif isinstance(f, BinaryOperator):
        return type(f)([_expand(subf) for subf in f.formulas])
    # elif isinstance(f, LDLfLogicalTrue):
    #     return PLTrue()
    # elif isinstance(f, LDLfLogicalFalse):
    #     return PLFalse()
    else:
        return f

LDLfAnd.Dual = LDLfOr
LDLfOr.Dual = LDLfAnd

LDLfDiamond.Dual = LDLfBox
LDLfBox.Dual = LDLfDiamond



from flloat.base.Symbol import Symbol
from flloat.parser.pl import PLParser
from flloat.semantics.pl import PLInterpretation
from flloat.syntax.pl import PLAnd, PLAtomic, PLNot, PLEquivalence, PLOr, PLImplies, PLFalse, PLTrue


def test_truth():
    sa, sb = Symbol("a"), Symbol("b")
    a, b = PLAtomic(sa), PLAtomic(sb)
    i_ = PLInterpretation(set())
    i_a = PLInterpretation({sa})
    i_b = PLInterpretation({sb})
    i_ab = PLInterpretation({sa, sb})

    a_and_b = PLAnd([a, b])
    assert not a_and_b.truth(i_)
    assert not a_and_b.truth(i_a)
    assert not a_and_b.truth(i_b)
    assert     a_and_b.truth(i_ab)

    not_a_and_not_b = PLAnd([PLNot(a), PLNot(b)])
    assert      not_a_and_not_b.truth(i_)
    assert not  not_a_and_not_b.truth(i_a)
    assert not  not_a_and_not_b.truth(i_b)
    assert not  not_a_and_not_b.truth(i_ab)

    material_implication = PLEquivalence([
        PLOr([PLNot(a), b]),
        PLNot(PLAnd([a, PLNot(b)])),
        PLImplies([a, b])
    ])

    # the equivalence is valid (i.e. satisfied for every interpretation)
    assert material_implication.truth(i_)
    assert material_implication.truth(i_a)
    assert material_implication.truth(i_b)
    assert material_implication.truth(i_ab)

    a_and_false_and_true = PLAnd([a, PLFalse(), PLTrue()])
    assert not a_and_false_and_true.truth(i_)
    assert not a_and_false_and_true.truth(i_a)
    assert not a_and_false_and_true.truth(i_b)
    assert not a_and_false_and_true.truth(i_ab)

    a_or_false_or_true = PLOr([a, PLFalse(), PLTrue()])
    assert a_or_false_or_true.truth(i_)
    assert a_or_false_or_true.truth(i_a)
    assert a_or_false_or_true.truth(i_b)
    assert a_or_false_or_true.truth(i_ab)


def test_parser():
    parser = PLParser()
    sa, sb = Symbol("A"), Symbol("B")
    a, b = PLAtomic(sa), PLAtomic(sb)

    a_and_b = parser("A & B")
    true_a_and_b = PLAnd([a, b])
    assert a_and_b == true_a_and_b

    material_implication = parser("!A | B <-> !(A & !B) <-> A->B")
    true_material_implication = PLEquivalence([
        PLOr([PLNot(a), b]),
        PLNot(PLAnd([a, PLNot(b)])),
        PLImplies([a, b])
    ])

    assert material_implication == true_material_implication

    true_a_and_false_and_true = PLAnd([a, PLFalse(), PLTrue()])
    a_and_false_and_true = parser("A & false & true")

    assert a_and_false_and_true == true_a_and_false_and_true


def test_nnf():
    parser = PLParser()
    sa, sb = Symbol("A"), Symbol("B")
    a, b = PLAtomic(sa), PLAtomic(sb)
    i_ = PLInterpretation(set())
    i_a = PLInterpretation({sa})
    i_b = PLInterpretation({sb})
    i_ab = PLInterpretation({sa, sb})

    not_a_and_b = parser("!(A&B)")
    nnf_not_a_and_b = parser("!A | !B")
    assert not_a_and_b.to_nnf() == nnf_not_a_and_b
    assert nnf_not_a_and_b == nnf_not_a_and_b.to_nnf()

    dup = parser("!(A | A)")
    nnf_dup = dup.to_nnf()
    assert nnf_dup == PLNot(a)

    material_implication = parser("!A | B <-> !(A & !B) <-> A->B")
    nnf_material_implication = parser("((!A | B) & (!A | B) & (!A | B)) | ((A & !B) & (A & !B) & (A & !B))")
    nnf_m = material_implication.to_nnf()
    assert nnf_m == nnf_material_implication.to_nnf()

    assert nnf_m.truth(i_) == material_implication.truth(i_) == nnf_material_implication.truth(i_) == True
    assert nnf_m.truth(i_a) == material_implication.truth(i_a) == nnf_material_implication.truth(i_a) == True
    assert nnf_m.truth(i_b) == material_implication.truth(i_b) == nnf_material_implication.truth(i_b) == True
    assert nnf_m.truth(i_ab) == material_implication.truth(i_ab) == nnf_material_implication.truth(i_ab) == True

def test_find_labels():
    parser = PLParser()

    # complete formula
    f = "!A | B <-> !(A & !B) <-> A->B"
    formula = parser(f)
    assert formula.find_labels() == {Symbol("A"), Symbol("B")}

    # more than one character
    f = "!A & (!AB & !A0)"
    formula = parser(f)
    assert formula.find_labels() == {Symbol(c) for c in {"A", "AB", "A0"}}



from abc import abstractmethod

from flloat.base.Formula import Formula, BinaryOperator, CommutativeBinaryOperator
from flloat.base.misc import Delta
from flloat.base.nnf import NNF
from flloat.base.truths import Truth, ImpliesTruth, EquivalenceTruth
from flloat.semantics.pl import PLInterpretation


class ConvertibleFormula(Formula):
    @abstractmethod
    def _convert(self):
        raise NotImplementedError


class BaseConvertibleFormula(ConvertibleFormula, Truth, NNF):
    def truth(self, *args):
        return self._convert().truth(*args)

    def _to_nnf(self):
        return self._convert().to_nnf()

    def negate(self):
        return self._convert().negate()


class DeltaConvertibleFormula(BaseConvertibleFormula, Delta):
    def _delta(self, i:PLInterpretation, epsilon=False):
        return self._convert().delta(i, epsilon)


class ImpliesConvertible(ImpliesTruth, BaseConvertibleFormula, BinaryOperator):
    @property
    def And(self):
        raise NotImplementedError

    @property
    def Or(self):
        raise NotImplementedError

    @property
    def Not(self):
        raise NotImplementedError

    def _convert(self):
        fs = self.formulas
        if len(fs) > 2:
            a, b = self.And(fs[:-1]), fs[-1]
        else:
            a, b = fs
        res = self.Or([self.Not(a), b])
        return res

class ImpliesDeltaConvertible(ImpliesConvertible, DeltaConvertibleFormula):
    pass

class EquivalenceConvertible(EquivalenceTruth, BaseConvertibleFormula, CommutativeBinaryOperator):
    @property
    def And(self):
        raise NotImplementedError

    @property
    def Or(self):
        raise NotImplementedError

    @property
    def Not(self):
        raise NotImplementedError

    def _convert(self):
        fs = self.formulas
        pos = self.And(fs)
        neg = self.And([self.Not(f) for f in fs])

        res = self.Or([pos, neg])
        return res

class EquivalenceDeltaConvertible(EquivalenceConvertible, DeltaConvertibleFormula):
    pass



from typing import Set, FrozenSet

from flloat.base.Symbol import Symbol
from flloat.base.hashable import Hashable
from flloat.utils import _powerset, powerset, ObjFactory, ObjConstructor

class _Alphabet(Hashable):
    def __init__(self, symbols: FrozenSet[Symbol]):
        super().__init__()
        self.symbols = symbols
        self._powerset = None

    def _members(self):
        return self.symbols

    def powerset(self):
        if self._powerset is None:
            self._powerset = _Alphabet(powerset(self.symbols))
        return self._powerset


class AlphabetConstructor(ObjConstructor):
    def __call__(self, symbols: Set[Symbol]) -> _Alphabet:
        f_symbols = frozenset(symbols)
        return super().__call__(f_symbols)

    @classmethod
    def fromStrings(cls, symbol_strings: Set[str]):
        f_symbols = frozenset(Symbol(s) for s in symbol_strings)
        return alphabet_factory.new(f_symbols)

alphabet_factory = ObjFactory(_Alphabet)
Alphabet = AlphabetConstructor(alphabet_factory)



from abc import ABC, abstractmethod
from functools import lru_cache

from flloat.base.Formula import UnaryOperator, CommutativeBinaryOperator, AtomicFormula, BinaryOperator
from flloat.utils import MAX_CACHE_SIZE


class NNF(ABC):

    @lru_cache(maxsize=MAX_CACHE_SIZE)
    def to_nnf(self):
        return self._to_nnf()

    @abstractmethod
    def _to_nnf(self):
        raise NotImplementedError

    @abstractmethod
    def negate(self):
        raise NotImplementedError


class DualNNF(NNF):
    @property
    def Dual(self):
        raise NotImplementedError

    @Dual.setter
    def Dual(self, x):
        self.Dual = x

class NotNNF(UnaryOperator, NNF):
    def _to_nnf(self):
        if not isinstance(self.f, AtomicFormula):
            return self.f.negate().to_nnf()
        else:
            return self.f.negate()

    def negate(self):
        return self.f


class DualUnaryOperatorNNF(UnaryOperator, DualNNF):
    @property
    def Not(self):
        raise NotImplementedError

    @Not.setter
    def Not(self, x):
        self.Not = x

    def _to_nnf(self):
        return type(self)(self.f.to_nnf())

    def negate(self):
        return self.Dual(self.Not(self.f))

class DualBinaryOperatorNNF(BinaryOperator, DualNNF):

    def _to_nnf(self):
        childs = [child.to_nnf() for child in self.formulas]
        return type(self)(childs).simplify()

    def negate(self):
        childs = [child.negate().to_nnf() for child in self.formulas]
        return self.Dual(childs).simplify()


class DualCommutativeOperatorNNF(CommutativeBinaryOperator, DualBinaryOperatorNNF):
    pass
    # def _to_nnf(self):
    #     childs = [child.to_nnf() for child in self.members]
    #     return type(self)(childs).simplify()
    #
    # def negate(self):
    #     childs = [child.negate().to_nnf() for child in self.members]
    #     return self.Dual(childs)





from flloat.base.Symbol import Symbol
from flloat.base.Symbols import Symbols
from flloat.base.parsing import Lexer, Parser
from flloat.syntax.pl import PLNot, PLAtomic, PLOr, PLAnd, PLImplies, PLEquivalence, PLTrue, PLFalse
from flloat.utils import sym2regexp


class PLLexer(Lexer):

    def __init__(self):
        super().__init__()

    reserved = {
        Symbols.TRUE.value  : 'TRUE',
        Symbols.FALSE.value : 'FALSE',
    }

    # List of token names.   This is always required
    tokens = (
        'ATOM',
        'NOT',
        'AND',
        'OR',
        'IMPLIES',
        'EQUIVALENCE',
        'LPAREN',
        'RPAREN'
    ) + tuple(reserved.values())

    # Regular expression rules for simple tokens
    t_NOT          = sym2regexp(Symbols.NOT)
    t_AND          = sym2regexp(Symbols.AND)
    t_OR           = sym2regexp(Symbols.OR)
    t_IMPLIES      = sym2regexp(Symbols.IMPLIES)
    t_EQUIVALENCE  = sym2regexp(Symbols.EQUIVALENCE)
    t_LPAREN       = sym2regexp(Symbols.ROUND_BRACKET_LEFT)
    t_RPAREN       = sym2regexp(Symbols.ROUND_BRACKET_RIGHT)



    def t_ATOM(self, t):
        r'[a-zA-Z_][a-zA-Z_0-9]*'
        t.type = PLLexer.reserved.get(t.value, 'ATOM')  # Check for reserved words
        return t


class PLParser(Parser):

    def __init__(self):
        lexer = PLLexer()
        precedence = (
            ('left', 'EQUIVALENCE'),
            ('right', 'IMPLIES'),
            ('left', 'OR'),
            ('left', 'AND'),
            ('right', 'NOT'),
        )
        super().__init__("pl", lexer.tokens, lexer, precedence)

    def p_formula_atom(self, p):
        """formula : ATOM
                   | TRUE
                   | FALSE"""
        if p[1]==Symbols.TRUE.value:
            p[0] = PLTrue()
        elif p[1]==Symbols.FALSE.value:
            p[0] = PLFalse()
        else:
            p[0] = PLAtomic(Symbol(p[1]))

    def p_formula_not(self, p):
        'formula : NOT formula'
        p[0] = PLNot(p[2])

    def p_formula_or(self, p):
        'formula : formula OR formula'
        p[0] = PLOr([p[1], p[3]])

    def p_formula_and(self, p):
        'formula : formula AND formula'
        p[0] = PLAnd([p[1], p[3]])

    def p_formula_implies(self, p):
        'formula : formula IMPLIES formula'
        p[0] = PLImplies([p[1], p[3]])

    def p_formula_equivalence(self, p):
        'formula : formula EQUIVALENCE formula'
        p[0] = PLEquivalence([p[1], p[3]])

    def p_formula_expression(self, p):
        'formula : LPAREN formula RPAREN'
        p[0] = p[2]


if __name__ == '__main__':
    parser = PLParser()
    while True:
        try:
            s = input('calc > ')
        except EOFError:
            break
        if not s: continue
        result = parser(s)
        print(result)



from flloat.base.Symbol import Symbol
from flloat.base.Symbols import Symbols
from flloat.parser.ltlf import LTLfParser
from flloat.semantics.ldlf import FiniteTrace
from flloat.semantics.pl import PLFalseInterpretation, PLInterpretation
from flloat.syntax.ltlf import LTLfAtomic, LTLfAnd, LTLfEquivalence, LTLfOr, LTLfNot, LTLfImplies, LTLfEventually, \
    LTLfAlways, LTLfUntil, LTLfRelease, LTLfNext, LTLfWeakNext, LTLfTrue, LTLfFalse
from flloat.syntax.pl import PLAtomic, PLTrue, PLFalse, PLAnd, PLOr


def test_parser():
    parser = LTLfParser()

    a, b, c = [LTLfAtomic(Symbol(c)) for c in "ABC"]

    assert parser("!A | B <-> !(A & !B) <-> A->B") == LTLfEquivalence([
        LTLfOr([LTLfNot(a), b]),
        LTLfNot(LTLfAnd([a, LTLfNot(b)])),
        LTLfImplies([a, b])
    ])

    assert parser("(X A) & (WX !B)") == LTLfAnd([
        LTLfNext(a),
        LTLfWeakNext(LTLfNot(b))
    ])

    assert parser("(F (A&B)) <-> !(G (!A | !B) )") == LTLfEquivalence([
        LTLfEventually(LTLfAnd([a, b])),
        LTLfNot(LTLfAlways(LTLfOr([LTLfNot(a), LTLfNot(b)])))
    ])

    assert parser("(A U B U C) <-> !(!A R !B R !C)") == LTLfEquivalence([
        LTLfUntil([a, b, c]),
        LTLfNot(LTLfRelease([LTLfNot(a), LTLfNot(b), LTLfNot(c)]))
    ])



def test_truth():
    parser = LTLfParser()
    t = FiniteTrace.fromStringSets([
        {"A"},
        {"A"},
        {"B"},
        {"B"},
        {"C"},
        {"C"},
    ])

    # Next and Weak Next
    f = "X A"
    assert parser("X A").truth(t, 0)
    assert not parser("X A").truth(t, 1)
    assert not parser("WX A").truth(t, 1)
    assert parser("X B").truth(t, 1)
    assert parser("X C").truth(t, 4)
    # at the last step, Next != WeakNext
    assert not parser("X C").truth(t, 5)
    assert parser("WX C").truth(t, 5)

    # Until
    f = "A U B U C"
    assert parser(f).truth(t, 0)
    assert parser(f).truth(t, 2)
    assert parser(f).truth(t, 4)
    assert not parser(f).truth(t, 10)

    assert not parser("A U C").truth(t, 0)
    assert not parser("C U B").truth(t, 0)

    # Release - dual of Until
    f = "(!A R !B R !C)"
    assert not parser(f).truth(t, 0)
    assert not parser(f).truth(t, 2)
    assert not parser(f).truth(t, 4)
    assert     parser(f).truth(t, 10)

    assert not parser("A U C").truth(t, 0)
    assert not parser("C U B").truth(t, 0)

    # Eventually
    assert      parser("F C & !A & !B").truth(t, 0)
    assert not  parser("F A & B & C").truth(t, 0)
    assert      parser("F G C").truth(t, 0)
    assert not  parser("F G B").truth(t, 0)

    # Always
    assert parser("G A | B | C").truth(t, 0)
    assert parser("G F (C & !A & !B)").truth(t, 0)
    assert not parser("G C").truth(t, 0)
    assert parser("G C").truth(t, 4)
    assert parser("G C").truth(t, 10)
    assert parser("G F C").truth(t, 0)


def test_nnf():
    parser = LTLfParser()
    a, b, c = [LTLfAtomic(Symbol(c)) for c in "ABC"]

    f = parser("!(A & !B)")
    assert f.to_nnf() == LTLfOr([LTLfNot(a), b])

    f = parser("!(!A | B)")
    assert f.to_nnf() == LTLfAnd([a, LTLfNot(b)])

    f = parser("!( (A->B) <-> (!A | B))")
    assert f.to_nnf() == LTLfAnd([
        LTLfAnd([a, LTLfNot(b)]),
        LTLfOr([LTLfNot(a), b]),
    ])

    # Next and Weak Next
    f = parser("!(X (A & B))")
    assert f.to_nnf() == LTLfWeakNext(LTLfOr([LTLfNot(a), LTLfNot(b)]))

    f = parser("!(WX (A & B))")
    assert f.to_nnf() == LTLfNext(LTLfOr([LTLfNot(a), LTLfNot(b)]))

    # Eventually and Always
    f = parser("!(F (A | B))")
    assert f.to_nnf() == LTLfAlways(LTLfAnd([LTLfNot(a), LTLfNot(b)])).to_nnf()

    f = parser("!(F (A | B))")
    assert f.to_nnf() == LTLfAlways(LTLfAnd([LTLfNot(a), LTLfNot(b)])).to_nnf()
    f = parser("!(G (A | B))")
    assert f.to_nnf() == LTLfEventually(LTLfAnd([LTLfNot(a), LTLfNot(b)])).to_nnf()

    # Until and Release
    f = parser("!(A U B)")
    assert f.to_nnf() == LTLfRelease([LTLfNot(a), LTLfNot(b)])
    f = parser("!(A R B)")
    assert f.to_nnf() == LTLfUntil([LTLfNot(a), LTLfNot(b)])


def test_delta():
    parser = LTLfParser()
    sa, sb = Symbol("A"), Symbol("B")
    a, b = PLAtomic(sa), PLAtomic(sb)

    i_ = PLFalseInterpretation()
    i_a = PLInterpretation({sa})
    i_b = PLInterpretation({sb})
    i_ab = PLInterpretation({sa, sb})

    true = PLTrue()
    false = PLFalse()

    assert parser("A").delta(i_) == false
    assert parser("A").delta(i_a) == true
    assert parser("A").delta(i_b) == false
    assert parser("A").delta(i_ab) == true

    assert parser("!A").delta(i_) == true
    assert parser("!A").delta(i_a) == false
    assert parser("!A").delta(i_b) == true
    assert parser("!A").delta(i_ab) == false

    assert parser("A & B").delta(i_) ==   PLAnd([false, false])
    assert parser("A & B").delta(i_a) ==  PLAnd([true, false])
    assert parser("A & B").delta(i_b) ==  PLAnd([false, true])
    assert parser("A & B").delta(i_ab) == PLAnd([true, true])

    assert parser("A | B").delta(i_) ==   PLOr([false, false])
    assert parser("A | B").delta(i_a) ==  PLOr([true, false])
    assert parser("A | B").delta(i_b) ==  PLOr([false, true])
    assert parser("A | B").delta(i_ab) == PLOr([true, true])

    assert parser("X A").delta(i_)   ==   PLAnd([LTLfAtomic(sa), LTLfEventually(LTLfTrue()).to_nnf()])
    assert parser("X A").delta(i_a)  ==   PLAnd([LTLfAtomic(sa), LTLfEventually(LTLfTrue()).to_nnf()])
    assert parser("X A").delta(i_b)  ==   PLAnd([LTLfAtomic(sa), LTLfEventually(LTLfTrue()).to_nnf()])
    assert parser("X A").delta(i_ab) ==   PLAnd([LTLfAtomic(sa), LTLfEventually(LTLfTrue()).to_nnf()])
    assert parser("X A").delta(i_, epsilon=True) == false

    assert parser("F A").delta(i_a) ==    PLOr([true,  PLAnd([LTLfEventually(LTLfTrue()).to_nnf(), true, LTLfUntil([LTLfTrue(), LTLfAtomic(sa)])])])
    assert parser("F A").delta(i_) ==     PLOr([false, PLAnd([LTLfEventually(LTLfTrue()).to_nnf(), true, LTLfUntil([LTLfTrue(), LTLfAtomic(sa)])])])
    assert parser("F A").delta(i_a, epsilon=True) == false
    assert parser("F A").delta(i_, epsilon=True) == false

    assert parser("G A").delta(i_a) ==  PLAnd([true, PLOr([false, LTLfAlways(LTLfFalse()).to_nnf(), LTLfRelease([LTLfFalse(), LTLfAtomic(sa)])])])
    assert parser("G A").delta(i_a) ==  PLAnd([true, PLOr([false, LTLfAlways(LTLfFalse()).to_nnf(), LTLfRelease([LTLfFalse(), LTLfAtomic(sa)])])])
    assert parser("G A").delta(i_a, epsilon=True) == true
    assert parser("G A").delta(i_,  epsilon=True) == true

    assert parser("A U B").delta(i_a) == PLOr([
        false,
        PLAnd([
            true,
            LTLfUntil([LTLfAtomic(sa), LTLfAtomic(sb)]),
            LTLfEventually(LTLfTrue()).to_nnf()
        ])
    ])

    assert parser("A R B").delta(i_a) == PLAnd([
        false,
        PLOr([
            true,
            LTLfRelease([LTLfAtomic(sa), LTLfAtomic(sb)]),
            LTLfAlways(LTLfFalse()).to_nnf()
        ])
    ])


def _dfa_test(name, parser, string_formula, alphabet, test_function):
    """temporary function to easily test both the full DFA and the on-the-fly DFA"""
    ltlf = parser(string_formula)
    ldlf = ltlf.to_LDLf()

    # dfa = ltlf.to_automaton(alphabet, determinize=False, minimize=False)
    # dfa.to_dot("tests/automata/" + name + "_ltlf_no_min", title=name + "\n" + string_formula)
    dfa = ltlf.to_automaton(alphabet, determinize=True, minimize=True)
    # dfa.to_dot("tests/automata/" + name + "_ltlf", title=name+"\n"+string_formula)
    test_function(dfa)
    dfa = ltlf.to_automaton(alphabet, on_the_fly=True)
    test_function(dfa)

    # LDLf equivalent
    # print(str(ltlf), str(ldlf.to_nnf()))
    # dfa = ldlf.to_automaton(alphabet, determinize=True, minimize=False)
    # dfa.to_dot("tests/automata/" + name + "_ldlf_no_min", title=name + "\n" + string_formula)
    dfa = ldlf.to_automaton(alphabet, determinize=True, minimize=True)
    # dfa.to_dot("tests/automata/" + name + "_ldlf", title=name+"\n"+str(ldlf.to_nnf()))
    test_function(dfa)
    dfa = ldlf.to_automaton(alphabet, on_the_fly=True)
    test_function(dfa)

def test_to_automaton():
    parser = LTLfParser()
    a, b, c = Symbol("A"), Symbol("B"), Symbol("C")
    alphabet_abc = {a, b, c}

    i_ = PLInterpretation(set())
    i_a = PLInterpretation({a})
    i_b = PLInterpretation({b})
    i_ab = PLInterpretation({a, b})

    ##################################################################################
    f = "A"
    name = "atomic"

    def test_f(dfa):
        assert not dfa.word_acceptance([])
        assert not dfa.word_acceptance([i_])
        assert     dfa.word_acceptance([i_a])
        assert not dfa.word_acceptance([i_b])
        assert     dfa.word_acceptance([i_ab])
        assert not dfa.word_acceptance([i_,   i_])
        assert not dfa.word_acceptance([i_,   i_a])
        assert not dfa.word_acceptance([i_,   i_b])
        assert not dfa.word_acceptance([i_,   i_ab])
        assert     dfa.word_acceptance([i_a,  i_])
        assert     dfa.word_acceptance([i_a,  i_a])
        assert     dfa.word_acceptance([i_a,  i_b])
        assert     dfa.word_acceptance([i_a,  i_ab])
        assert not dfa.word_acceptance([i_b,  i_])
        assert not dfa.word_acceptance([i_b,  i_a])
        assert not dfa.word_acceptance([i_b,  i_b])
        assert not dfa.word_acceptance([i_b,  i_ab])
        assert     dfa.word_acceptance([i_ab, i_])
        assert     dfa.word_acceptance([i_ab, i_a])
        assert     dfa.word_acceptance([i_ab, i_b])
        assert     dfa.word_acceptance([i_ab, i_ab])

        assert dfa.word_acceptance([i_a, i_, i_ab, i_b])
        assert not dfa.word_acceptance([i_, i_ab])

    _dfa_test(name, parser, f, alphabet_abc, test_f)
    ##################################################################################

    ##################################################################################
    f = "X A"
    name = "next"

    def test_f(dfa):
        assert not dfa.word_acceptance([])
        assert not dfa.word_acceptance([i_])
        assert not dfa.word_acceptance([i_a])
        assert not dfa.word_acceptance([i_b])
        assert not dfa.word_acceptance([i_ab])
        assert not dfa.word_acceptance([i_, i_])
        assert     dfa.word_acceptance([i_, i_a])
        assert not dfa.word_acceptance([i_, i_b])
        assert     dfa.word_acceptance([i_, i_ab])
        assert not dfa.word_acceptance([i_a, i_])
        assert     dfa.word_acceptance([i_a, i_a])
        assert not dfa.word_acceptance([i_a, i_b])
        assert     dfa.word_acceptance([i_a, i_ab])
        assert not dfa.word_acceptance([i_b, i_])
        assert     dfa.word_acceptance([i_b, i_a])
        assert not dfa.word_acceptance([i_b, i_b])
        assert     dfa.word_acceptance([i_b, i_ab])
        assert not dfa.word_acceptance([i_ab, i_])
        assert     dfa.word_acceptance([i_ab, i_a])
        assert not dfa.word_acceptance([i_ab, i_b])
        assert     dfa.word_acceptance([i_ab, i_ab])

        assert not dfa.word_acceptance([i_a, i_b, i_])
        assert not dfa.word_acceptance([i_, i_, i_, i_, i_a, i_, i_ab, i_, i_])

    _dfa_test(name, parser, f, alphabet_abc, test_f)
    ##################################################################################

    ##################################################################################
    f = "WX A"
    name = "weak_next"

    def test_f(dfa):
        assert     dfa.word_acceptance([])
        assert     dfa.word_acceptance([i_])
        assert     dfa.word_acceptance([i_a])
        assert     dfa.word_acceptance([i_b])
        assert     dfa.word_acceptance([i_ab])
        assert not dfa.word_acceptance([i_, i_])
        assert     dfa.word_acceptance([i_, i_a])
        assert not dfa.word_acceptance([i_, i_b])
        assert     dfa.word_acceptance([i_, i_ab])
        assert not dfa.word_acceptance([i_a, i_])
        assert     dfa.word_acceptance([i_a, i_a])
        assert not dfa.word_acceptance([i_a, i_b])
        assert     dfa.word_acceptance([i_a, i_ab])
        assert not dfa.word_acceptance([i_b, i_])
        assert     dfa.word_acceptance([i_b, i_a])
        assert not dfa.word_acceptance([i_b, i_b])
        assert     dfa.word_acceptance([i_b, i_ab])
        assert not dfa.word_acceptance([i_ab, i_])
        assert     dfa.word_acceptance([i_ab, i_a])
        assert not dfa.word_acceptance([i_ab, i_b])
        assert     dfa.word_acceptance([i_ab, i_ab])


        assert     dfa.word_acceptance([i_b])
        assert not dfa.word_acceptance([i_b, i_b, i_b])
        assert     dfa.word_acceptance([i_b, i_a, i_ab])

    _dfa_test(name, parser, f, alphabet_abc, test_f)
    ##################################################################################

    ##################################################################################
    f = "A U B"
    name = "until"

    def test_f(dfa):
        assert not dfa.word_acceptance([])
        assert not dfa.word_acceptance([i_])
        assert not dfa.word_acceptance([i_a])
        assert     dfa.word_acceptance([i_b])
        assert     dfa.word_acceptance([i_ab])
        assert not dfa.word_acceptance([i_, i_])
        assert not dfa.word_acceptance([i_, i_a])
        assert not dfa.word_acceptance([i_, i_b])
        assert not dfa.word_acceptance([i_, i_ab])
        assert not dfa.word_acceptance([i_a, i_])
        assert not dfa.word_acceptance([i_a, i_a])
        assert     dfa.word_acceptance([i_a, i_b])
        assert     dfa.word_acceptance([i_a, i_ab])
        assert     dfa.word_acceptance([i_b, i_])
        assert     dfa.word_acceptance([i_b, i_a])
        assert     dfa.word_acceptance([i_b, i_b])
        assert     dfa.word_acceptance([i_b, i_ab])
        assert     dfa.word_acceptance([i_ab, i_])
        assert     dfa.word_acceptance([i_ab, i_a])
        assert     dfa.word_acceptance([i_ab, i_b])
        assert     dfa.word_acceptance([i_ab, i_ab])

    _dfa_test(name, parser, f, alphabet_abc, test_f)
    ##################################################################################

    #################################################################################
    f = "!A R !B"
    name = "release"

    def test_f(dfa):
        assert     dfa.word_acceptance([])
        assert     dfa.word_acceptance([i_])
        assert     dfa.word_acceptance([i_a])
        assert not dfa.word_acceptance([i_b])
        assert not dfa.word_acceptance([i_ab])
        assert     dfa.word_acceptance([i_, i_])
        assert     dfa.word_acceptance([i_, i_a])
        assert     dfa.word_acceptance([i_, i_b])
        assert     dfa.word_acceptance([i_, i_ab])
        assert     dfa.word_acceptance([i_a, i_])
        assert     dfa.word_acceptance([i_a, i_a])
        assert not dfa.word_acceptance([i_a, i_b])
        assert not dfa.word_acceptance([i_a, i_ab])
        assert not dfa.word_acceptance([i_b, i_])
        assert not dfa.word_acceptance([i_b, i_a])
        assert not dfa.word_acceptance([i_b, i_b])
        assert not dfa.word_acceptance([i_b, i_ab])
        assert not dfa.word_acceptance([i_ab, i_])
        assert not dfa.word_acceptance([i_ab, i_a])
        assert not dfa.word_acceptance([i_ab, i_b])
        assert not dfa.word_acceptance([i_ab, i_ab])

    _dfa_test(name, parser, f, alphabet_abc, test_f)
    #################################################################################

    ##################################################################################
    f = "F A"
    name = "eventually"

    def test_f(dfa):
        assert not dfa.word_acceptance([])
        assert not dfa.word_acceptance([i_])
        assert     dfa.word_acceptance([i_a])
        assert not dfa.word_acceptance([i_b])
        assert     dfa.word_acceptance([i_ab])
        assert not dfa.word_acceptance([i_, i_])
        assert     dfa.word_acceptance([i_, i_a])
        assert not dfa.word_acceptance([i_, i_b])
        assert     dfa.word_acceptance([i_, i_ab])
        assert     dfa.word_acceptance([i_a, i_])
        assert     dfa.word_acceptance([i_a, i_a])
        assert     dfa.word_acceptance([i_a, i_b])
        assert     dfa.word_acceptance([i_a, i_ab])
        assert not dfa.word_acceptance([i_b, i_])
        assert     dfa.word_acceptance([i_b, i_a])
        assert not dfa.word_acceptance([i_b, i_b])
        assert     dfa.word_acceptance([i_b, i_ab])
        assert     dfa.word_acceptance([i_ab, i_])
        assert     dfa.word_acceptance([i_ab, i_a])
        assert     dfa.word_acceptance([i_ab, i_b])
        assert     dfa.word_acceptance([i_ab, i_ab])

        assert not dfa.word_acceptance([i_b, i_b, i_b])
        assert     dfa.word_acceptance([i_b, i_a, i_ab])

    _dfa_test(name, parser, f, alphabet_abc, test_f)
    ##################################################################################

    ##################################################################################
    f = "G A"
    name = "always"

    def test_f(dfa):
        assert     dfa.word_acceptance([])
        assert not dfa.word_acceptance([i_])
        assert     dfa.word_acceptance([i_a])
        assert not dfa.word_acceptance([i_b])
        assert     dfa.word_acceptance([i_ab])
        assert not dfa.word_acceptance([i_, i_])
        assert not dfa.word_acceptance([i_, i_a])
        assert not dfa.word_acceptance([i_, i_b])
        assert not dfa.word_acceptance([i_, i_ab])
        assert not dfa.word_acceptance([i_a, i_])
        assert     dfa.word_acceptance([i_a, i_a])
        assert not dfa.word_acceptance([i_a, i_b])
        assert     dfa.word_acceptance([i_a, i_ab])
        assert not dfa.word_acceptance([i_b, i_])
        assert not dfa.word_acceptance([i_b, i_a])
        assert not dfa.word_acceptance([i_b, i_b])
        assert not dfa.word_acceptance([i_b, i_ab])
        assert not dfa.word_acceptance([i_ab, i_])
        assert     dfa.word_acceptance([i_ab, i_a])
        assert not dfa.word_acceptance([i_ab, i_b])
        assert     dfa.word_acceptance([i_ab, i_ab])

        assert not dfa.word_acceptance([i_b, i_b, i_b])
        assert not dfa.word_acceptance([i_b, i_a, i_ab])
        assert     dfa.word_acceptance([i_a, i_a, i_ab])
        assert not dfa.word_acceptance([i_a, i_a, i_ab, i_b])
        assert     dfa.word_acceptance([i_a, i_a, i_ab, i_a])

    _dfa_test(name, parser, f, alphabet_abc, test_f)
    ##################################################################################

    ##################################################################################
    f = "G (A -> X(B) )"
    name = "always implication and next"

    def test_f(dfa):
        assert     dfa.word_acceptance([])
        assert     dfa.word_acceptance([i_])
        assert not dfa.word_acceptance([i_a])
        assert     dfa.word_acceptance([i_b])
        assert not dfa.word_acceptance([i_ab])
        assert     dfa.word_acceptance([i_, i_])
        assert not dfa.word_acceptance([i_, i_a])
        assert     dfa.word_acceptance([i_, i_b])
        assert not dfa.word_acceptance([i_, i_ab])
        assert not dfa.word_acceptance([i_a, i_])
        assert not dfa.word_acceptance([i_a, i_a])
        assert     dfa.word_acceptance([i_a, i_b])
        assert not dfa.word_acceptance([i_a, i_ab])
        assert     dfa.word_acceptance([i_b, i_])
        assert not dfa.word_acceptance([i_b, i_a])
        assert     dfa.word_acceptance([i_b, i_b])
        assert not dfa.word_acceptance([i_b, i_ab])
        assert not dfa.word_acceptance([i_ab, i_])
        assert not dfa.word_acceptance([i_ab, i_a])
        assert     dfa.word_acceptance([i_ab, i_b])
        assert not dfa.word_acceptance([i_ab, i_ab])

        assert     dfa.word_acceptance([i_b, i_b, i_b])
        assert not dfa.word_acceptance([i_b, i_a, i_ab])
        assert not dfa.word_acceptance([i_a, i_a, i_ab])
        assert not dfa.word_acceptance([i_a, i_a, i_ab, i_b])
        assert     dfa.word_acceptance([i_a, i_ab, i_ab, i_b])
        assert not dfa.word_acceptance([i_a, i_ab, i_ab, i_])
        # very important
        assert     dfa.word_acceptance([i_a, i_ab, i_ab, i_b, i_])
        assert     dfa.word_acceptance([i_a, i_ab, i_ab, i_b, i_b])
        assert     dfa.word_acceptance([i_a, i_ab, i_b, i_ab, i_b])

    _dfa_test(name, parser, f, alphabet_abc, test_f)
    ##################################################################################

    ##################################################################################
    f = "G (A -> X(G(B)) )"
    name = "always implication next always"

    def test_f(dfa):
        assert     dfa.word_acceptance([])
        assert     dfa.word_acceptance([i_])
        assert not dfa.word_acceptance([i_a])
        assert     dfa.word_acceptance([i_b])
        assert not dfa.word_acceptance([i_ab])
        assert     dfa.word_acceptance([i_, i_])
        assert not dfa.word_acceptance([i_, i_a])
        assert     dfa.word_acceptance([i_, i_b])
        assert not dfa.word_acceptance([i_, i_ab])
        assert not dfa.word_acceptance([i_a, i_])
        assert not dfa.word_acceptance([i_a, i_a])
        assert     dfa.word_acceptance([i_a, i_b])
        assert not dfa.word_acceptance([i_a, i_ab])
        assert     dfa.word_acceptance([i_b, i_])
        assert not dfa.word_acceptance([i_b, i_a])
        assert     dfa.word_acceptance([i_b, i_b])
        assert not dfa.word_acceptance([i_b, i_ab])
        assert not dfa.word_acceptance([i_ab, i_])
        assert not dfa.word_acceptance([i_ab, i_a])
        assert     dfa.word_acceptance([i_ab, i_b])
        assert not dfa.word_acceptance([i_ab, i_ab])

        assert     dfa.word_acceptance([i_b, i_b, i_b])
        assert not dfa.word_acceptance([i_b, i_a, i_ab])
        assert not dfa.word_acceptance([i_a, i_a, i_ab])
        assert not dfa.word_acceptance([i_a, i_a, i_ab, i_b])
        assert     dfa.word_acceptance([i_a, i_ab, i_ab, i_b])
        assert not dfa.word_acceptance([i_a, i_ab, i_ab, i_])
        # very important
        assert not dfa.word_acceptance([i_a, i_ab, i_ab, i_b, i_])
        assert     dfa.word_acceptance([i_a, i_ab, i_ab, i_b, i_b])
        assert     dfa.word_acceptance([i_a, i_ab, i_b, i_ab, i_b])

    _dfa_test(name, parser, f, alphabet_abc, test_f)

# def test_sapientino_formula():
#     p = LTLfParser()
#
#     f = '''(!bip U ( (A | B | C) & bip))
#                 & (G (A & bip -> X(G (bip -> !A) ) ) )
#                 & (G (B & bip -> X(G (bip -> !B) ) ) )
#                 & (G (C & bip -> X(G (bip -> !C) ) ) )
#                 & (
#                       (G (A & bip -> X( !bip U ((B | C) & bip) ) ) )
#                     | (G (B & bip -> X( !bip U ((A | C) & bip) ) ) )
#                     | (G (C & bip -> X( !bip U ((A | B) & bip) ) ) )
#                 )'''
#
#     f = '''( !bip U A & bip )'''
#
#     a, b, c, bip = Symbol("A"), Symbol("B"), Symbol("C"), Symbol("bip")
#     alphabet_abc = {a, b, c, bip}
#
#     def tr(l):
#         return FiniteTrace.fromSymbolSets(l).trace
#
#
#     def test_sapientino(dfa):
#         dfa.to_dot("temptemptemp")
#         assert not dfa.word_acceptance([])
#         assert not dfa.word_acceptance(tr([{bip}]))
#         assert     dfa.word_acceptance(tr([{bip, a}]))
#         assert     dfa.word_acceptance(tr([{bip, b}]))
#         assert     dfa.word_acceptance(tr([{bip, c}]))
#         assert not dfa.word_acceptance(tr([{bip, a}, {bip}]))
#         assert not dfa.word_acceptance(tr([{bip, b}, {bip}]))
#         assert not dfa.word_acceptance(tr([{bip, c}, {bip}]))
#     _dfa_test(p, f, alphabet_abc, test_sapientino)



from abc import abstractmethod, ABC
from typing import Sequence, Set

from flloat.base.Symbol import Symbol
from flloat.base.Symbols import Symbols
from flloat.base.hashable import Hashable


class Formula(Hashable):
    def __init__(self):
        super().__init__()

    @abstractmethod
    def find_labels(self) -> Set[Symbol]:
        return set()

    def simplify(self):
        return self

class AtomicFormula(Formula):
    def __init__(self, s:Symbol):
        super().__init__()
        self.s = s

    def _members(self):
        return self.s

    def __str__(self):
        return str(self.s)

    def find_labels(self):
        return {self.s}

class Operator(Formula):
    base_expression = Symbols.ROUND_BRACKET_LEFT.value + "%s" + Symbols.ROUND_BRACKET_RIGHT.value

    @property
    def operator_symbol(self) -> str:
        raise NotImplementedError


class UnaryOperator(Operator):
    def __init__(self, f: Formula):
        super().__init__()
        self.f = f.simplify()

    def __str__(self):
        return self.operator_symbol + Symbols.ROUND_BRACKET_LEFT.value + str(self.f) + Symbols.ROUND_BRACKET_RIGHT.value

    def _members(self):
        return (self.operator_symbol, self.f)

    def __lt__(self, other):
        return self.f.__lt__(other.f)

    def find_labels(self):
        return self.f.find_labels()


OperatorChilds = Sequence[Formula]
CommOperatorChilds = Set[Formula]


class BinaryOperator(Operator):
    """A generic binary formula"""


    def __init__(self, formulas:OperatorChilds):
        super().__init__()
        assert len(formulas) >= 2
        self.formulas = tuple(formulas)
        self.formulas = self._popup()

    def __str__(self):
        return "(" + (" "+self.operator_symbol+" ").join(map(str,self.formulas)) + ")"

    def _members(self):
        return (self.operator_symbol, self.formulas)

    def _popup(self):
        """recursively find commutative binary operator
        among child formulas and pop up them at the same level"""
        res = ()
        for child in self.formulas:
            if type(child) == type(self):
                superchilds = child.formulas
                res += superchilds
            else:
                res += (child, )
        return tuple(res)

    def find_labels(self):
        return set.union(*map(lambda f: f.find_labels(), self.formulas))

class CommutativeBinaryOperator(BinaryOperator):
    """A generic commutative binary formula"""


    def __init__(self, formulas:OperatorChilds, idempotence=True):
        # Assuming idempotence: e.g. A & A === A
        super().__init__(formulas)
        self.idempotence = idempotence
        if idempotence:
            # order does not matter -> set operation
            # remove duplicates -> set operation
            self.formulas_set = frozenset(self.formulas)
            # unique representation -> sorting
            self.members = tuple(sorted(self.formulas_set, key=lambda x: hash(x)))

    def simplify(self):
        if self.idempotence:
            if len(self.formulas_set) == 1:
                return next(iter(self.formulas_set)).simplify()
            else:
                return type(self)(self.members)
        else:
            return self


    def _members(self):
        if self.idempotence:
            return (self.operator_symbol, self.members)
            # return (self.operator_symbol, self.formulas_set)
        else:
            return super()._members()

    def find_labels(self):
        return set.union(*map(lambda f: f.find_labels(), self.formulas))

    def __str__(self):
        if self.idempotence:
            return "(" + (" "+self.operator_symbol+" ").join(map(str,self.members)) + ")"
        else:
            return super().__str__()



from abc import ABC, abstractmethod

from flloat.base.Formula import Formula, UnaryOperator, BinaryOperator
from flloat.base.Symbols import Symbols


class Truth(ABC):
    @abstractmethod
    def truth(self, *args) -> bool:
        raise NotImplementedError

class NotTruth(UnaryOperator, Truth):
    def __init__(self, f):
        UnaryOperator.__init__(self, f)

    operator_symbol = Symbols.NOT.value
    def truth(self, *args):
        return not self.f.truth(*args)


class AndTruth(BinaryOperator, Truth):
    operator_symbol = Symbols.AND.value
    def truth(self, *args):
        for f in self.formulas:
            if not f.truth(*args):
                return False
        return True
        # return all(f.truth(*args) for f in self.formulas)

class OrTruth(BinaryOperator, Truth):
    operator_symbol = Symbols.OR.value
    def truth(self, *args):
        for f in self.formulas:
            if f.truth(*args):
                return True
        return False
        # return any(f.truth(*args) for f in self.formulas)


class ImpliesTruth(BinaryOperator, Truth):
    operator_symbol = Symbols.IMPLIES.value


    def truth(self, *args):
        # eq_formula = self._convert()
        # return eq_formula.truth(i)
        fs = self.formulas
        N = len(fs)
        evaluation = lambda x, y: not x or y
        truth = evaluation(fs[0].truth(*args), fs[1].truth(*args))
        for idx in range(2, N):
            # if at some point the premise is False, then return True
            if not truth:
                return True
            else:
                truth = evaluation(truth, fs[idx].truth(*args))

        return truth

class EquivalenceTruth(BinaryOperator, Truth):
    operator_symbol = Symbols.EQUIVALENCE.value

    def truth(self, *args):
        fs = self.formulas
        N = len(fs)
        t = [f.truth(*args) for f in fs]
        # either all true or all false
        return all(t) or not any(t)



from abc import ABC, abstractmethod
from copy import copy

class Hashable(ABC):

    def __init__(self):
        self._hash = None

    @abstractmethod
    def _members(self):
        raise NotImplementedError

    def __eq__(self, other):
        if type(other) is type(self):
            return self._members() == other._members()
        else:
            return False

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(self._members())
        return self._hash

    def __getstate__(self):
        d = copy(self.__dict__)
        d.pop("_hash")
        return d

    def __setstate__(self, state):
        self.__dict__ = state
        self._hash = None



from itertools import chain, combinations
# import pythomata.base.DFA
# import pythomata.base.NFA
#
# # https://docs.python.org/3/library/itertools.html#recipes
# from pythogic.base.Alphabet_ import Alphabet_
#
from typing import Set, FrozenSet

from flloat.base.Symbols import Symbols


def powerset(iterable) -> FrozenSet:
    "powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"
    combs = _powerset(iterable)
    res = frozenset(frozenset(x) for x in combs)
    # res = map(frozenset, combs)
    return res

def _powerset(iterable):
    s = list(set(iterable))
    combs = chain.from_iterable(combinations(s, r) for r in range(len(s) + 1))
    for c in combs:
        yield c

def sym2regexp(sym:Symbols):
    s = sym.value
    if s in r"|()+?*.[]":
        return r"\%s"%s
    else:
        return s

MAX_CACHE_SIZE = 1024


class ObjFactory(object):
    def __init__(self, cls):
        self.cls = cls
        self.objects = {}

    def new(self, *args, **kwargs):
        old = self.objects.get(args, None)
        if old is not None:
            return old
        else:
            new = self.cls(*args, **kwargs)
            # notice: we do not save kwargs, so it works only with positional arguments.
            self.objects[args] = new
            return new


class ObjConstructor(object):
    def __init__(self, obj_factory:ObjFactory):
        self.obj_factory = obj_factory

    def __call__(self, *args, **kwargs):
        return self.obj_factory.new(*args, **kwargs)




from typing import FrozenSet, Set, List

from flloat.base.Interpretation import Interpretation
from flloat.base.Symbol import Symbol
from flloat.utils import ObjFactory, ObjConstructor


class _PLInterpretation(Interpretation):

    def __init__(self, true_propositions:FrozenSet[Symbol]):
        super().__init__()
        self.true_propositions = frozenset(true_propositions)

    def _members(self):
        # return tuple(sorted(self.true_propositions, key=lambda x: x.name))
        return self.true_propositions

    def __contains__(self, item:Symbol):
        return item in self.true_propositions

    def __iter__(self):
        return self.true_propositions.__iter__()

    def __str__(self):
        return "{" + ", ".join(map(str,self._members())) + "}"

    def __repr__(self):
        return self.__str__()

class PLTrueInterpretation(_PLInterpretation):
    def __init__(self):
        super().__init__(frozenset())

    def _members(self):
        return PLTrueInterpretation

    def __contains__(self, item):
        return True

class PLFalseInterpretation(_PLInterpretation):
    def __init__(self):
        super().__init__(frozenset())

    def __contains__(self, item):
        return False

class _PLInterpretationConstructor(ObjConstructor):
    def __call__(self, true_propositions:Set[Symbol]):
        f_sym = frozenset(true_propositions)
        return super().__call__(f_sym)

    def fromStrings(self, strings:List[str]):
        return self(set(Symbol(s) for s in strings))



plinterpretation_factory = ObjFactory(_PLInterpretation)
PLInterpretation = _PLInterpretationConstructor(plinterpretation_factory)



# -*- coding: utf-8 -*-
from flloat.base.Symbols import Symbols, ALL_SYMBOLS
from flloat.base.hashable import Hashable


class Symbol(str, Hashable):
    """A class to represent a symbol (actually, a wrap for a string)"""
    def __init__(self, name:str):
        str.__init__(self)
        Hashable.__init__(self)
        self.name = name

    def _members(self):
        return self.name

    # def __hash__(self):
    #     return hash(self._members())
    #
    # def __eq__(self, other):
    #     if type(other) is type(self):
    #         return self._members() == other._members()
    #     else:
    #         return False




class FunctionSymbol(Symbol):
    """A class to represent a function symbol ("""
    def __init__(self, name: str, arity: int):
        super().__init__(name)
        self.arity = arity

    def __str__(self):
        return self.name + Symbols.CARET.value + str(self.arity)

    def _members(self):
        return (self.name, self.arity)


class ConstantSymbol(FunctionSymbol):
    def __init__(self, name:str):
        super().__init__(name, 0)


class PredicateSymbol(Symbol):
    def __init__(self, name: str, arity):
        super().__init__(name)
        self.arity = arity

    def __str__(self):
        return self.name + "^" + str(self.arity)

    def __repr__(self):
        return self.name + "^" + str(self.arity)

    def _members(self):
        return (self.name, self.arity)


class TrueSymbol(Symbol):
    def __init__(self):
        super().__init__(Symbols.TOP.value)

class FalseSymbol(Symbol):
    def __init__(self):
        super().__init__(Symbols.BOTTOM.value)

class LastSymbol(Symbol):
    def __init__(self):
        super().__init__(Symbols.LAST.value)



"""a temporary python script for small tests"""
from flloat.base.Symbol import Symbol
from flloat.parser.ldlf import LDLfParser
from flloat.parser.ltlf import LTLfParser
from flloat.semantics.ldlf import FiniteTrace


def k(n):
    from flloat.base.Symbol import Symbol
    from flloat.parser.ldlf import LDLfParser
    parser = LDLfParser()
    cc = ["c%d" % d for d in range(n)]
    labels = [Symbol(d) for d in cc]
    f = "<%s>tt" % ";".join(cc)
    print(f)

    f = parser(f)
    nfa = f.to_automaton(labels)
    nfa.to_dot("temp_nfa.NFA")
    print("nfa")
    dfa = nfa.determinize()
    dfa.to_dot("temp_det.DFA")
    print("dfa")
    dfa = dfa.minimize().trim()
    dfa.to_dot("temp_min.DFA")
    print("dfatrim")

def breakout_formula():
    p = LDLfParser()
    f = p("<(!l0 & !l1 & !l2)*;(l0 & !l1 & !l2);(l0 & !l1 & !l2)*;(l0 & l1 & !l2); (l0 & l1 & !l2)*; l0 & l1 & l2>tt")
    f.to_automaton(determinize=True).to_dot("temp")

def reward_shaping_example_formula():
    p = LDLfParser()
    a, b, c = [Symbol(c) for c in "ABC"]
    f = p("<A*; B>tt")
    f.to_automaton({a, b, c}, determinize=True).complete().to_dot("temp.DFA")
    f.to_automaton({a, b, c}).to_dot("temp.NFA")
    k = f.to_automaton({a, b, c}, on_the_fly=True)

    t_false = FiniteTrace.fromStringSets([
        {"A"},
        {"A"},
        {"C"},
        {"B"}
    ])

    t_true = FiniteTrace.fromStringSets([
        {"A"},
        {"A"},
        {"B"},
        {"B"}
    ])

    # k.word_acceptance(t_false.trace)
    # k.word_acceptance(t_true.trace)

    print(k.cur_state)
    for s in t_true.trace:
        k.make_transition(s)
        print(k.cur_state)

def sapientino_test():
    p = LTLfParser()
    #  | cellC0_1 | cellC0_2
    f = "(!b U (cellC0_0 & b)) & G (cellC0_0 & b -> X(G b -> !cellC0_0))"
    # "G (cellC0_1 & b -> X(G b -> !cellC0_1))"

    # (!b U (cellC0_0 & b)) &
    f = "G (cellC0_0 & b -> X(G (b -> !cellC0_0)) )"
    # f = "G (A -> X(G (B -> C)) )"
    f = "G ( A -> X(G(B)) )"
    # f = "G (A -> X(B) )"
    # f = "X(F A)"
    f = '''(!b U ( (A | B | C) & b))
                & (G (A & b -> X(G (b -> !A) ) ) )
                & (G (B & b -> X(G (b -> !B) ) ) )
                & (G (C & b -> X(G (b -> !C) ) ) )
                & (
                      (G (A & b -> X( !b U ((B | C) & b) ) ) )
                    | (G (B & b -> X( !b U ((A | C) & b) ) ) )
                    | (G (C & b -> X( !b U ((A | B) & b) ) ) )
                )'''

    # f = '''(!b U ( A & b & !END))
    #                 & (G (A & b -> X(G (b -> !A) ) ) )
    #                 & (G (A & b -> X( !b U END & b ) ) )'''

    # f = '''(!b U ( (A | B) & b))
    #                 & (G (A & b -> X(G (b -> !A) ) ) )
    #                 & (G (B & b -> X(G (b -> !B) ) ) )
    #                 & (
    #                       (G (A & b -> X( !b U ( B & b) ) ) )
    #                     | (G (B & b -> X( !b U ( A & b) ) ) )
    #                 )'''

    print(f)
    # f = '''(!b U ( (A | B) & b ))
    #             & (G (A & b -> X( G (b -> !A) ) ) )
    #             & (G (B & b -> X( G (b -> !B) ) ) )'''

    # f = '''(!b U A & b | B & b)
    #             & G (b & (A|B))
    #             & (F (A & b)) & (F (B & b))
    #             & (G (A & b -> X( G (b -> !A) ) ) )
    #             & (G (B & b -> X( G (b -> !B) ) ) )'''

    # & (G(A & b -> X( !b U B & b) ) )
    # & (G(B & b -> X( !b U A & b) ) )


    # print(p(f))
    # print(p(f).to_automaton().to_dot("nfa"))

    p(f).to_automaton(determinize=True, minimize=True).map_to_int().to_dot("TEMP.ltlf")

    # nfa = p(f).to_automaton(determinize=False, minimize=False)
    # nfa.map_to_int().to_dot("OLD_TEMP_NFA.ltlf.nfa")
    # dfa = nfa.determinize().trim().minimize().trim()
    # dfa.map_to_int().to_dot("OLD_TEMP_MIN_TRIM_DFA.ltlf.dfa")

    # dfa = p(f).to_automaton(minimize=False)
    # dfa = dfa.map_states_and_action(states_map={v: k for k, v in enumerate(dfa.states)})
    # dfa.to_dot("TEMP.ltlf")

    # p = LDLfParser()
    # p("[true](<A>tt)").to_automaton(determinize=True).to_dot("temptemp.ldlf")

def ldlf_ltlf():
    d = LDLfParser()
    # d("<(!b)*; A & b & !(B | C | D | E | F); "
    #    "(!b)*; B & b & !(A | C | D | E | F); "
    #    "(!b)*; C & b & !(B | A | D | E | F); "+
    #    "(!b)*; D & b & !(B | C | A | E | F); "
    #    "(!b)*; E & b & !(B | C | D | A | F); "
    #    "(!b)*; F & b & !(B | C | D | E | A)>tt").to_automaton(determinize=True).to_dot("temptemp.ldlf")

    # d("<A;B;C;D;E;F;G;H;I;L;M;N;O;P>tt").to_automaton(determinize=True, minimize=False)
        # .to_dot("temptemp.ldlf")

    d("<(!bip)*;red & bip;(!bip)*;green & bip;(!bip)*;blue & bip;(!bip)*;pink & bip;(!bip)*;brown & bip;(!bip)*;gray & bip;(!bip)*;purple & bip;(!bip)*;no_color & bip>tt") \
        .to_automaton(determinize=True, minimize=True).to_dot("temptemp.ldlf")

    # d("[((<true>tt)?;true)*]([true]ff | <A>tt)").to_automaton(determinize=False, minimize=False).to_dot("temptemp.ldlf.1")
    # d("[((<true>tt)?;true)*]([true]ff | <A>tt)").to_automaton(determinize=True, minimize=False).to_dot("temptemp.ldlf.2")
    # d("[((<true>tt)?;true)*]([true]ff | <A>tt)").to_automaton(determinize=True, minimize=True).to_dot("temptemp.ldlf.3")
    # d("[true*]([true]ff | <A>tt)").to_automaton(determinize=True, minimize=True).to_dot("temptemp.ldlf")

    # d("<((<true>tt)?;true)*>(<true>tt & <A>tt)").to_automaton(determinize=True).to_dot("temptemp.ldlf")
    # d("<true*>(<true>tt & <A>tt)").to_automaton(determinize=True).to_dot("temptemp.ldlf")

    # ltlfp = LTLfParser()
    # f = "G ( A -> X(G(B)) )"
    # print(ltlfp(f).to_LDLf().to_nnf())
    # ltlfp(f).to_LDLf().to_automaton(determinize=True, minimize=True).to_dot("temptemp.ltlf")
    # ltlfp("WX AAAA").to_LDLf().to_automaton(determinize=True, minimize=True).to_dot("temptemp.ltlf")


def sapientino_ldlf():
    # f = "<(!bip)*;red & bip;(!bip)*;green & bip;(!bip)*;blue & bip;(!bip)*;pink & bip;(!bip)*;brown & bip;(!bip)*;gray & bip;(!bip)*;purple & bip>tt"
    # d = LDLfParser()
    # parsed = d(f)
    # parsed.to_automaton(determinize=True, minimize=True).to_dot("sapientino_ldlf_1")

    # f = "<true*;red & bip;true*;green & bip;true*;blue & bip;true*;pink & bip;true*;brown & bip;true*;gray & bip;true*;purple & bip>tt & " \
    #     " [((<true>tt)?;true)*](<bip -> (red | green | blue | pink | brown | gray | purple)>tt  | [true]ff)"
    f = " [true*](<bip -> (A | C)>tt  | [true]ff)"

    f = "<(!bip & !(red | green | blue | pink | brown | gray | purple))*;" \
        "red & bip & !(green | blue | pink | brown | gray | purple);" \
        "(!bip & !(red | green | blue | pink | brown | gray | purple))*;" \
        "green & bip & !(red | blue | pink | brown | gray | purple);" \
        "(!bip & !(red | green | blue | pink | brown | gray | purple))*;" \
        "blue & bip & !(red | green | pink | brown | gray | purple);" \
        "(!bip & !(red | green | blue | pink | brown | gray | purple))*;" \
        "pink & bip & !(red | green | blue | brown | gray | purple);" \
        "(!bip & !(red | green | blue | pink | brown | gray | purple))*;" \
        "brown & bip & !(red | green | blue | pink | gray | purple);" \
        "(!bip & !(red | green | blue | pink | brown | gray | purple))*;" \
        "gray & bip & !(red | green | blue | pink | brown | purple);" \
        "(!bip & !(red | green | blue | pink | brown | gray | purple))*;" \
        "purple & bip & !(red | green | blue | pink | brown | gray)" \
        ">end"

    # f = "<(!bip & !red & !green & !blue & !pink & !brown & !gray & !purple)*; red>tt"
    d = LDLfParser()
    parsed = d(f)

    parsed.to_automaton(determinize=True, minimize=True).to_dot("sapientino_ldlf_2")


def minecraft_ldlf():
    # f = "<(!get & !use)*;get & wood; (!get & !use)*;use & workbench; (!get & !use)*;get & iron; (!get & !use)*; use & toolshed>tt"
    f = "<true*><true;A;B>tt"
    # f = "[true*](<!get & !use>tt | [true]ff)"
    # f = "[true*](<(get | use) -> (wood | grass | iron | gold | gem | toolshed | workbench | factory) >tt | [true]ff)"
    f = "<(!l0 & !l1 & !l2)*;(l0 & !l1 & !l2);(l0 & !l1 & !l2)*;(l0 & l1 & !l2); (l0 & l1 & !l2)*; l0 & l1 & l2>tt"




    d = LDLfParser()
    # f = "[(true)*]((<A>(tt) | [true](ff)))"
    # f = "<true*;A;true*;B>tt | <A & B>tt"
    # f = "<true*>(<true*;A;A*;(A & B);(A & B)*;(A & B & C)>tt)"
    f = "<true*>(<true*;A;A*;(A & B);(A & B)*;(A & B & C)>tt)"
    f = "<true*>(<true*;A;A*;(A & B);(A & B)*;C>tt)"
    f = "<true*>(<true*;(get_wood);(get_wood)*;use_workbench>tt)"
    f = "<true*>(<true*;(get_iron);(get_iron)*;(get_iron & get_wood);(get_iron & get_wood)*;use_factory>tt)"
    f = "<true*>(<(!get_wood & !use_factory)*;(!get_wood & !use_factory & get_iron);(!get_wood & !use_factory & get_iron)*;" \
        "(get_iron & get_wood & !use_factory);(get_iron & get_wood & !use_factory)*;use_factory>tt)"

    f = "<true*>(<(!get_iron & !get_wood & !use_factory)*;(get_iron & !get_wood & !use_factory);(get_iron & !get_wood & !use_factory)*;(get_iron & get_wood & !use_factory);(get_iron & get_wood & !use_factory)*;use_factory>tt)"
    # f = "<true*>(<true*;(get_wood);(get_wood)*;(get_wood & use_workbench)>tt)"

    # f = "<true*>(<true*;(get_wood);(get_wood)*;(get_wood & use_toolshed);(get_wood & use_toolshed)*;(get_wood & use_toolshed & get_grass);(get_wood & use_toolshed & get_grass)*;(get_wood & use_toolshed & get_grass & use_workbench)>tt)"
    # d = LTLfParser()
    # f = "G(A)"
    # print(d(f).to_LDLf().to_nnf())
    # f = "F(A & F(B))"
    # f = "F(A & F(B)) & F(C & F(B))"
    # f = "F(A & F(B)) & F(C & F(B))"
    # f = "F(A & F(C)) & F(B & F(C))"
    # f = "F(A & F(B & F(C)))"
    # f = "F(A & F(B & X(F(C))))"
    # f = "F(A & F(B)) & G(A-> F(B))"
    # f = "F(A & (A U B))"
    # f = "F(got_wood & F(used_toolshed & F(got_grass & F(used_workbench))))"
    # parsed = d(f)
    # print(str(parsed.to_LDLf()))

    parsed = d(f)
    parsed.to_automaton(determinize=True, minimize=True)._numbering_states().to_dot("prove")


def thesis():
    # p = LTLfParser()
    # p("G A").to_automaton(determinize=True, minimize=False).to_dot("temptemp-dfa.ltlf")
    # p("F A").to_automaton(determinize=True, minimize=False).to_dot("temptemp-dfa.ltlf")
    # p("F true").to_automaton(determinize=True, minimize=False).to_dot("temptemp-dfa.ltlf")
    # p("F( (A & !B) & F(!A & B))").to_automaton(determinize=True).complete().to_dot("2event")

    p = LDLfParser()
    # f = p("<(A & !B & !C & !D)*;(!A & B & !C & !D)*;(!A & !B & C & !D)*; (!A & !B & !C & D)>end")
    f = p("<(!A & !B & !C)*;(A & !B & !C);(A & !B & !C)*;(!A & B & !C);(!A & B & !C)*;(!A & !B & C);(!A & !B & C)*>end")
    f.to_automaton(determinize=True).to_dot("abc")


if __name__ == '__main__':
    pass
    # k(4)
    #

    # sapientino_test()
    # ldlf_ltlf()
    # sapientino_ldlf()
    # minecraft_ldlf()
    thesis()

